# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import ast
import logging

import mccabe

from pylsp import hookimpl, lsp

log = logging.getLogger(__name__)

THRESHOLD = "threshold"
DEFAULT_THRESHOLD = 15


@hookimpl
def pylsp_lint(config, workspace, document):
    with workspace.report_progress("lint: mccabe"):
        threshold = config.plugin_settings("mccabe", document_path=document.path).get(
            THRESHOLD, DEFAULT_THRESHOLD
        )
        log.debug("Running mccabe lint with threshold: %s", threshold)

        try:
            tree = compile(document.source, document.path, "exec", ast.PyCF_ONLY_AST)
        except SyntaxError:
            # We'll let the other linters point this one out
            return None

        visitor = mccabe.PathGraphingAstVisitor()
        visitor.preorder(tree, visitor)

        diags = []
        for graph in visitor.graphs.values():
            if graph.complexity() >= threshold:
                diags.append(
                    {
                        "source": "mccabe",
                        "range": {
                            "start": {
                                "line": graph.lineno - 1,
                                "character": graph.column,
                            },
                            "end": {
                                "line": graph.lineno - 1,
                                "character": len(document.lines[graph.lineno]),
                            },
                        },
                        "message": "Cyclomatic complexity too high: %s (threshold %s)"
                        % (graph.complexity(), threshold),
                        "severity": lsp.DiagnosticSeverity.Warning,
                    }
                )

        return diags



from __future__ import annotations

import logging
from collections.abc import Sequence
from typing import Callable, ContextManager

from rope.base.taskhandle import BaseJobSet, BaseTaskHandle

from pylsp._utils import throttle
from pylsp.workspace import Workspace

log = logging.getLogger(__name__)
Report = Callable[[str, int], None]


class PylspJobSet(BaseJobSet):
    count: int = 0
    done: int = 0
    _reporter: Report
    _report_iter: ContextManager
    job_name: str = ""

    def __init__(self, count: int | None, report_iter: ContextManager) -> None:
        if count is not None:
            self.count = count
        self._reporter = report_iter.__enter__()
        self._report_iter = report_iter

    def started_job(self, name: str | None) -> None:
        if name:
            self.job_name = name

    def finished_job(self) -> None:
        self.done += 1
        if self.get_percent_done() is not None and int(self.get_percent_done()) >= 100:
            if self._report_iter is None:
                return
            self._report_iter.__exit__(None, None, None)
            self._report_iter = None
        else:
            self._report()

    def check_status(self) -> None:
        pass

    def get_percent_done(self) -> float | None:
        if self.count == 0:
            return 0
        return (self.done / self.count) * 100

    def increment(self) -> None:
        """
        Increment the number of tasks to complete.

        This is used if the number is not known ahead of time.
        """
        self.count += 1
        self._report()

    @throttle(0.5)
    def _report(self) -> None:
        percent = int(self.get_percent_done())
        message = f"{self.job_name} {self.done}/{self.count}"
        log.debug(f"Reporting {message} {percent}%")
        self._reporter(message, percent)


class PylspTaskHandle(BaseTaskHandle):
    name: str
    observers: list
    job_sets: list[PylspJobSet]
    stopped: bool
    workspace: Workspace
    _report: Callable[[str, str], None]

    def __init__(self, workspace: Workspace) -> None:
        self.workspace = workspace
        self.job_sets = []
        self.observers = []

    def create_jobset(self, name="JobSet", count: int | None = None):
        report_iter = self.workspace.report_progress(
            name, None, None, skip_token_initialization=True
        )
        result = PylspJobSet(count, report_iter)
        self.job_sets.append(result)
        self._inform_observers()
        return result

    def stop(self) -> None:
        pass

    def current_jobset(self) -> BaseJobSet | None:
        pass

    def add_observer(self) -> None:
        pass

    def is_stopped(self) -> bool:
        pass

    def get_jobsets(self) -> Sequence[BaseJobSet]:
        pass

    def _inform_observers(self) -> None:
        for observer in self.observers:
            observer()



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging

from pylsp import hookimpl

log = logging.getLogger(__name__)

MODULES = [
    "OpenGL",
    "PIL",
    "array",
    "audioop",
    "binascii",
    "cPickle",
    "cStringIO",
    "cmath",
    "collections",
    "datetime",
    "errno",
    "exceptions",
    "gc",
    "imageop",
    "imp",
    "itertools",
    "marshal",
    "math",
    "matplotlib",
    "mmap",
    "mpmath",
    "msvcrt",
    "networkx",
    "nose",
    "nt",
    "numpy",
    "operator",
    "os",
    "os.path",
    "pandas",
    "parser",
    "rgbimg",
    "scipy",
    "signal",
    "skimage",
    "sklearn",
    "statsmodels",
    "strop",
    "sympy",
    "sys",
    "thread",
    "time",
    "wx",
    "xxsubtype",
    "zipimport",
    "zlib",
]


@hookimpl
def pylsp_settings():
    # Setup default modules to preload, and rope extension modules
    return {
        "plugins": {"preload": {"modules": MODULES}},
        "rope": {"extensionModules": MODULES},
    }


@hookimpl
def pylsp_initialize(config) -> None:
    for mod_name in config.plugin_settings("preload").get("modules", []):
        try:
            __import__(mod_name)
            log.debug("Preloaded module %s", mod_name)
        except Exception:
            # Catch any exception since not only ImportError can be raised here
            # For example, old versions of NumPy can cause a ValueError.
            # See spyder-ide/spyder#13985
            pass



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import functools
import inspect
import logging
import os
import pathlib
import re
import subprocess
import sys
import threading
import time
from typing import Optional

import docstring_to_markdown
import jedi

JEDI_VERSION = jedi.__version__

# Eol chars accepted by the LSP protocol
# the ordering affects performance
EOL_CHARS = ["\r\n", "\r", "\n"]
EOL_REGEX = re.compile(f"({'|'.join(EOL_CHARS)})")

log = logging.getLogger(__name__)


def debounce(interval_s, keyed_by=None):
    """Debounce calls to this function until interval_s seconds have passed."""

    def wrapper(func):
        timers = {}
        lock = threading.Lock()

        @functools.wraps(func)
        def debounced(*args, **kwargs):
            sig = inspect.signature(func)
            call_args = sig.bind(*args, **kwargs)
            key = call_args.arguments[keyed_by] if keyed_by else None

            def run():
                with lock:
                    del timers[key]
                return func(*args, **kwargs)

            with lock:
                old_timer = timers.get(key)
                if old_timer:
                    old_timer.cancel()

                timer = threading.Timer(interval_s, run)
                timers[key] = timer
                timer.start()

        return debounced

    return wrapper


def throttle(seconds=1):
    """Throttles calls to a function every `seconds` seconds."""

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not hasattr(wrapper, "last_call"):
                wrapper.last_call = 0
            if time.time() - wrapper.last_call >= seconds:
                wrapper.last_call = time.time()
                return func(*args, **kwargs)

        return wrapper

    return decorator


def find_parents(root, path, names):
    """Find files matching the given names relative to the given path.

    Args:
        path (str): The file path to start searching up from.
        names (list[str]): The file/directory names to look for.
        root (str): The directory at which to stop recursing upwards.

    Note:
        The path MUST be within the root.
    """
    if not root:
        return []

    if not os.path.commonprefix((root, path)):
        log.warning("Path %r not in %r", path, root)
        return []

    # Split the relative by directory, generate all the parent directories, then check each of them.
    # This avoids running a loop that has different base-cases for unix/windows
    # e.g. /a/b and /a/b/c/d/e.py -> ['/a/b', 'c', 'd']
    dirs = [root] + os.path.relpath(os.path.dirname(path), root).split(os.path.sep)

    # Search each of /a/b/c, /a/b, /a
    while dirs:
        search_dir = os.path.join(*dirs)
        existing = list(
            filter(os.path.exists, [os.path.join(search_dir, n) for n in names])
        )
        if existing:
            return existing
        dirs.pop()

    # Otherwise nothing
    return []


def path_to_dot_name(path):
    """Given a path to a module, derive its dot-separated full name."""
    directory = os.path.dirname(path)
    module_name, _ = os.path.splitext(os.path.basename(path))
    full_name = [module_name]
    while os.path.exists(os.path.join(directory, "__init__.py")):
        this_directory = os.path.basename(directory)
        directory = os.path.dirname(directory)
        full_name = [this_directory] + full_name
    return ".".join(full_name)


def match_uri_to_workspace(uri, workspaces):
    if uri is None:
        return None
    max_len, chosen_workspace = -1, None
    path = pathlib.Path(uri).parts
    for workspace in workspaces:
        workspace_parts = pathlib.Path(workspace).parts
        if len(workspace_parts) > len(path):
            continue
        match_len = 0
        for workspace_part, path_part in zip(workspace_parts, path):
            if workspace_part == path_part:
                match_len += 1
        if match_len > 0:
            if match_len > max_len:
                max_len = match_len
                chosen_workspace = workspace
    return chosen_workspace


def list_to_string(value):
    return ",".join(value) if isinstance(value, list) else value


def merge_dicts(dict_a, dict_b):
    """Recursively merge dictionary b into dictionary a.

    If override_nones is True, then
    """

    def _merge_dicts_(a, b):
        for key in set(a.keys()).union(b.keys()):
            if key in a and key in b:
                if isinstance(a[key], dict) and isinstance(b[key], dict):
                    yield (key, dict(_merge_dicts_(a[key], b[key])))
                elif isinstance(a[key], list) and isinstance(b[key], list):
                    yield (key, list(set(a[key] + b[key])))
                elif b[key] is not None:
                    yield (key, b[key])
                else:
                    yield (key, a[key])
            elif key in a:
                yield (key, a[key])
            elif b[key] is not None:
                yield (key, b[key])

    return dict(_merge_dicts_(dict_a, dict_b))


def escape_plain_text(contents: str) -> str:
    """
    Format plain text to display nicely in environments which do not respect whitespaces.
    """
    contents = contents.replace("\t", "\u00a0" * 4)
    contents = contents.replace("  ", "\u00a0" * 2)
    return contents


def escape_markdown(contents: str) -> str:
    """
    Format plain text to display nicely in Markdown environment.
    """
    # escape markdown syntax
    contents = re.sub(r"([\\*_#[\]])", r"\\\1", contents)
    # preserve white space characters
    contents = escape_plain_text(contents)
    return contents


def wrap_signature(signature):
    return "```python\n" + signature + "\n```\n"


SERVER_SUPPORTED_MARKUP_KINDS = {"markdown", "plaintext"}


def choose_markup_kind(client_supported_markup_kinds: list[str]):
    """Choose a markup kind supported by both client and the server.

    This gives priority to the markup kinds provided earlier on the client preference list.
    """
    for kind in client_supported_markup_kinds:
        if kind in SERVER_SUPPORTED_MARKUP_KINDS:
            return kind
    return "markdown"


class Formatter:
    command: list[str]

    @property
    def is_installed(self) -> bool:
        """Returns whether formatter is available"""
        if not hasattr(self, "_is_installed"):
            self._is_installed = self._is_available_via_cli()
        return self._is_installed

    def format(self, code: str, line_length: int) -> str:
        """Formats code"""
        return subprocess.check_output(
            [
                sys.executable,
                "-m",
                *self.command,
                "--line-length",
                str(line_length),
                "-",
            ],
            input=code,
            text=True,
        ).strip()

    def _is_available_via_cli(self) -> bool:
        try:
            subprocess.check_output(
                [
                    sys.executable,
                    "-m",
                    *self.command,
                    "--help",
                ],
            )
            return True
        except subprocess.CalledProcessError:
            return False


class RuffFormatter(Formatter):
    command = ["ruff", "format"]


class BlackFormatter(Formatter):
    command = ["black"]


formatters = {"ruff": RuffFormatter(), "black": BlackFormatter()}


def format_signature(signature: str, config: dict, signature_formatter: str) -> str:
    """Formats signature using ruff or black if either is available."""
    as_func = f"def {signature.strip()}:\n    pass"
    line_length = config.get("line_length", 88)
    formatter = formatters[signature_formatter]
    if formatter.is_installed:
        try:
            return (
                formatter.format(as_func, line_length=line_length)
                .removeprefix("def ")
                .removesuffix(":\n    pass")
            )
        except subprocess.CalledProcessError as e:
            log.warning("Signature formatter failed %s", e)
    else:
        log.warning(
            "Formatter %s was requested but it does not appear to be installed",
            signature_formatter,
        )
    return signature


def convert_signatures_to_markdown(signatures: list[str], config: dict) -> str:
    signature_formatter = config.get("formatter", "black")
    if signature_formatter:
        signatures = [
            format_signature(
                signature, signature_formatter=signature_formatter, config=config
            )
            for signature in signatures
        ]
    return wrap_signature("\n".join(signatures))


def format_docstring(
    contents: str,
    markup_kind: str,
    signatures: Optional[list[str]] = None,
    signature_config: Optional[dict] = None,
):
    """Transform the provided docstring into a MarkupContent object.

    If `markup_kind` is 'markdown' the docstring will get converted to
    markdown representation using `docstring-to-markdown`; if it is
    `plaintext`, it will be returned as plain text.
    Call signatures of functions (or equivalent code summaries)
    provided in optional `signatures` argument will be prepended
    to the provided contents of the docstring if given.
    """
    if not isinstance(contents, str):
        contents = ""

    if markup_kind == "markdown":
        wrapped_signatures = convert_signatures_to_markdown(
            signatures if signatures is not None else [], config=signature_config or {}
        )

        if contents != "":
            try:
                value = docstring_to_markdown.convert(contents)
            except docstring_to_markdown.UnknownFormatError:
                # try to escape the Markdown syntax instead:
                value = escape_markdown(contents)

            if signatures:
                value = wrapped_signatures + "\n\n" + value
        else:
            value = contents

            if signatures:
                value = wrapped_signatures

        return {"kind": "markdown", "value": value}
    value = contents
    if signatures:
        value = "\n".join(signatures) + "\n\n" + value
    return {"kind": "plaintext", "value": escape_plain_text(value)}


def clip_column(column, lines, line_number):
    """
    Normalise the position as per the LSP that accepts character positions > line length

    https://microsoft.github.io/language-server-protocol/specification#position
    """
    max_column = (
        len(lines[line_number].rstrip("\r\n")) if len(lines) > line_number else 0
    )
    return min(column, max_column)


def position_to_jedi_linecolumn(document, position):
    """
    Convert the LSP format 'line', 'character' to Jedi's 'line', 'column'

    https://microsoft.github.io/language-server-protocol/specification#position
    """
    code_position = {}
    if position:
        code_position = {
            "line": position["line"] + 1,
            "column": clip_column(
                position["character"], document.lines, position["line"]
            ),
        }
    return code_position


if os.name == "nt":
    import ctypes

    kernel32 = ctypes.windll.kernel32
    PROCESS_QUERY_INFROMATION = 0x1000

    def is_process_alive(pid):
        """Check whether the process with the given pid is still alive.

        Running `os.kill()` on Windows always exits the process, so it can't be used to check for an alive process.
        see: https://docs.python.org/3/library/os.html?highlight=os%20kill#os.kill

        Hence ctypes is used to check for the process directly via windows API avoiding any other 3rd-party dependency.

        Args:
            pid (int): process ID

        Returns:
            bool: False if the process is not alive or don't have permission to check, True otherwise.
        """
        process = kernel32.OpenProcess(PROCESS_QUERY_INFROMATION, 0, pid)
        if process != 0:
            kernel32.CloseHandle(process)
            return True
        return False

else:
    import errno

    def is_process_alive(pid):
        """Check whether the process with the given pid is still alive.

        Args:
            pid (int): process ID

        Returns:
            bool: False if the process is not alive or don't have permission to check, True otherwise.
        """
        if pid < 0:
            return False
        try:
            os.kill(pid, 0)
        except OSError as e:
            return e.errno == errno.EPERM
        return True


def get_eol_chars(text):
    """Get EOL chars used in text."""
    match = EOL_REGEX.search(text)
    if match:
        return match.group(0)
    return None



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

import jedi

from pylsp import _utils, hookimpl, uris

if TYPE_CHECKING:
    from jedi.api import Script
    from jedi.api.classes import Name

    from pylsp.config.config import Config
    from pylsp.workspace import Document

log = logging.getLogger(__name__)


MAX_JEDI_GOTO_HOPS = 100


def _resolve_definition(
    maybe_defn: Name, script: Script, settings: dict[str, Any]
) -> Name:
    for _ in range(MAX_JEDI_GOTO_HOPS):
        if maybe_defn.is_definition() or maybe_defn.module_path != script.path:
            break
        defns = script.goto(
            follow_imports=settings.get("follow_imports", True),
            follow_builtin_imports=settings.get("follow_builtin_imports", True),
            line=maybe_defn.line,
            column=maybe_defn.column,
        )
        if len(defns) == 1:
            maybe_defn = defns[0]
        else:
            break
    return maybe_defn


@hookimpl
def pylsp_definitions(
    config: Config, document: Document, position: dict[str, int]
) -> list[dict[str, Any]]:
    settings = config.plugin_settings("jedi_definition")
    code_position = _utils.position_to_jedi_linecolumn(document, position)
    script = document.jedi_script(use_document_path=True)
    auto_import_modules = jedi.settings.auto_import_modules

    try:
        jedi.settings.auto_import_modules = []
        definitions = script.goto(
            follow_imports=settings.get("follow_imports", True),
            follow_builtin_imports=settings.get("follow_builtin_imports", True),
            **code_position,
        )
        definitions = [_resolve_definition(d, script, settings) for d in definitions]
    finally:
        jedi.settings.auto_import_modules = auto_import_modules

    follow_builtin_defns = settings.get("follow_builtin_definitions", True)
    return [
        {
            "uri": uris.uri_with(document.uri, path=str(d.module_path)),
            "range": {
                "start": {"line": d.line - 1, "character": d.column},
                "end": {"line": d.line - 1, "character": d.column + len(d.name)},
            },
        }
        for d in definitions
        if d.is_definition() and (follow_builtin_defns or _not_internal_definition(d))
    ]


def _not_internal_definition(definition: Name) -> bool:
    return (
        definition.line is not None
        and definition.column is not None
        and definition.module_path is not None
        and not definition.in_builtin_module()
    )



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging
import os

import parso

from pylsp import _utils, hookimpl, lsp
from pylsp.plugins._resolvers import LABEL_RESOLVER, SNIPPET_RESOLVER

log = logging.getLogger(__name__)

# Map to the LSP type
# > Valid values for type are ``module``, `` class ``, ``instance``, ``function``,
# > ``param``, ``path``, ``keyword``, ``property`` and ``statement``.
# see: https://jedi.readthedocs.io/en/latest/docs/api-classes.html#jedi.api.classes.BaseName.type
_TYPE_MAP = {
    "module": lsp.CompletionItemKind.Module,
    "namespace": lsp.CompletionItemKind.Module,  # to be added in Jedi 0.18+
    "class": lsp.CompletionItemKind.Class,
    "instance": lsp.CompletionItemKind.Reference,
    "function": lsp.CompletionItemKind.Function,
    "param": lsp.CompletionItemKind.Variable,
    "path": lsp.CompletionItemKind.File,
    "keyword": lsp.CompletionItemKind.Keyword,
    "property": lsp.CompletionItemKind.Property,  # added in Jedi 0.18
    "statement": lsp.CompletionItemKind.Variable,
}

# Types of parso nodes for which snippet is not included in the completion
_IMPORTS = ("import_name", "import_from")

# Types of parso node for errors
_ERRORS = ("error_node",)


@hookimpl
def pylsp_completions(config, document, position):
    """Get formatted completions for current code position"""
    settings = config.plugin_settings("jedi_completion", document_path=document.path)
    resolve_eagerly = settings.get("eager", False)
    signature_config = config.settings().get("signature", {})

    code_position = _utils.position_to_jedi_linecolumn(document, position)
    code_position["fuzzy"] = settings.get("fuzzy", False)
    completions = document.jedi_script(use_document_path=True).complete(**code_position)

    if not completions:
        return None

    completion_capabilities = config.capabilities.get("textDocument", {}).get(
        "completion", {}
    )
    item_capabilities = completion_capabilities.get("completionItem", {})
    snippet_support = item_capabilities.get("snippetSupport")
    supported_markup_kinds = item_capabilities.get("documentationFormat", ["markdown"])
    preferred_markup_kind = _utils.choose_markup_kind(supported_markup_kinds)

    should_include_params = settings.get("include_params")
    should_include_class_objects = settings.get("include_class_objects", False)
    should_include_function_objects = settings.get("include_function_objects", False)

    max_to_resolve = settings.get("resolve_at_most", 25)
    modules_to_cache_for = settings.get("cache_for", None)
    if modules_to_cache_for is not None:
        LABEL_RESOLVER.cached_modules = modules_to_cache_for
        SNIPPET_RESOLVER.cached_modules = modules_to_cache_for

    include_params = (
        snippet_support and should_include_params and use_snippets(document, position)
    )
    include_class_objects = (
        snippet_support
        and should_include_class_objects
        and use_snippets(document, position)
    )
    include_function_objects = (
        snippet_support
        and should_include_function_objects
        and use_snippets(document, position)
    )

    ready_completions = [
        _format_completion(
            c,
            markup_kind=preferred_markup_kind,
            include_params=include_params if c.type in ["class", "function"] else False,
            resolve=resolve_eagerly,
            resolve_label_or_snippet=(i < max_to_resolve),
            snippet_support=snippet_support,
            signature_config=signature_config,
        )
        for i, c in enumerate(completions)
    ]

    # TODO split up once other improvements are merged
    if include_class_objects:
        for i, c in enumerate(completions):
            if c.type == "class":
                completion_dict = _format_completion(
                    c,
                    markup_kind=preferred_markup_kind,
                    include_params=False,
                    resolve=resolve_eagerly,
                    resolve_label_or_snippet=(i < max_to_resolve),
                    snippet_support=snippet_support,
                    signature_config=signature_config,
                )
                completion_dict["kind"] = lsp.CompletionItemKind.TypeParameter
                completion_dict["label"] += " object"
                ready_completions.append(completion_dict)

    if include_function_objects:
        for i, c in enumerate(completions):
            if c.type == "function":
                completion_dict = _format_completion(
                    c,
                    markup_kind=preferred_markup_kind,
                    include_params=False,
                    resolve=resolve_eagerly,
                    resolve_label_or_snippet=(i < max_to_resolve),
                    snippet_support=snippet_support,
                    signature_config=signature_config,
                )
                completion_dict["kind"] = lsp.CompletionItemKind.TypeParameter
                completion_dict["label"] += " object"
                ready_completions.append(completion_dict)

    for completion_dict in ready_completions:
        completion_dict["data"] = {"doc_uri": document.uri}

    # most recently retrieved completion items, used for resolution
    document.shared_data["LAST_JEDI_COMPLETIONS"] = {
        # label is the only required property; here it is assumed to be unique
        completion["label"]: (completion, data)
        for completion, data in zip(ready_completions, completions)
    }

    return ready_completions or None


@hookimpl
def pylsp_completion_item_resolve(
    config,
    completion_item,
    document,
):
    """Resolve formatted completion for given non-resolved completion"""
    shared_data = document.shared_data["LAST_JEDI_COMPLETIONS"].get(
        completion_item["label"]
    )

    completion_capabilities = config.capabilities.get("textDocument", {}).get(
        "completion", {}
    )
    item_capabilities = completion_capabilities.get("completionItem", {})
    supported_markup_kinds = item_capabilities.get("documentationFormat", ["markdown"])
    preferred_markup_kind = _utils.choose_markup_kind(supported_markup_kinds)

    if shared_data:
        completion, data = shared_data
        return _resolve_completion(
            completion,
            data,
            markup_kind=preferred_markup_kind,
            signature_config=config.settings().get("signature", {}),
        )
    return completion_item


def is_exception_class(name):
    """
    Determine if a class name is an instance of an Exception.

    This returns `False` if the name given corresponds with a instance of
    the 'Exception' class, `True` otherwise
    """
    try:
        return name in [cls.__name__ for cls in Exception.__subclasses__()]
    except AttributeError:
        # Needed in case a class don't uses new-style
        # class definition in Python 2
        return False


def use_snippets(document, position):
    """
    Determine if it's necessary to return snippets in code completions.

    This returns `False` if a completion is being requested on an import
    statement, `True` otherwise.
    """
    line = position["line"]
    lines = document.source.split("\n", line)
    act_lines = [lines[line][: position["character"]]]
    line -= 1
    last_character = ""
    while line > -1:
        act_line = lines[line]
        if (
            act_line.rstrip().endswith("\\")
            or act_line.rstrip().endswith("(")
            or act_line.rstrip().endswith(",")
        ):
            act_lines.insert(0, act_line)
            line -= 1
            if act_line.rstrip().endswith("("):
                # Needs to be added to the end of the code before parsing
                # to make it valid, otherwise the node type could end
                # being an 'error_node' for multi-line imports that use '('
                last_character = ")"
        else:
            break
    if "(" in act_lines[-1].strip():
        last_character = ")"
    code = "\n".join(act_lines).rsplit(";", maxsplit=1)[-1].strip() + last_character
    tokens = parso.parse(code)
    expr_type = tokens.children[0].type
    return expr_type not in _IMPORTS and not (expr_type in _ERRORS and "import" in code)


def _resolve_completion(completion, d, markup_kind: str, signature_config: dict):
    completion["detail"] = _detail(d)
    try:
        docs = _utils.format_docstring(
            d.docstring(raw=True),
            signatures=[signature.to_string() for signature in d.get_signatures()],
            markup_kind=markup_kind,
            signature_config=signature_config,
        )
    except Exception:
        docs = ""
    completion["documentation"] = docs
    return completion


def _format_completion(
    d,
    markup_kind: str,
    include_params=True,
    resolve=False,
    resolve_label_or_snippet=False,
    snippet_support=False,
    signature_config=None,
):
    completion = {
        "label": _label(d, resolve_label_or_snippet),
        "kind": _TYPE_MAP.get(d.type),
        "sortText": _sort_text(d),
        "insertText": d.name,
    }

    if resolve:
        completion = _resolve_completion(
            completion, d, markup_kind, signature_config=signature_config
        )

    # Adjustments for file completions
    if d.type == "path":
        path = os.path.normpath(d.name)

        # If the completion ends with os.sep, it means it's a directory. So we add os.sep at the end
        # to ease additional file completions.
        if d.name.endswith(os.sep):
            if os.name == "nt":
                path = path + "\\"
            else:
                path = path + "/"

        # Escape to prevent conflicts with the code snippets grammer
        # See also https://github.com/python-lsp/python-lsp-server/issues/373
        if snippet_support:
            path = path.replace("\\", "\\\\")
            path = path.replace("/", "\\/")

        completion["insertText"] = path

    if include_params and not is_exception_class(d.name):
        snippet = _snippet(d, resolve_label_or_snippet)
        completion.update(snippet)

    return completion


def _label(definition, resolve=False):
    if not resolve:
        return definition.name
    sig = LABEL_RESOLVER.get_or_create(definition)
    if sig:
        return sig
    return definition.name


def _snippet(definition, resolve=False):
    if not resolve:
        return {}
    snippet = SNIPPET_RESOLVER.get_or_create(definition)
    return snippet


def _detail(definition):
    try:
        return definition.parent().full_name or ""
    except AttributeError:
        return definition.full_name or ""


def _sort_text(definition):
    """Ensure builtins appear at the bottom.
    Description is of format <type>: <module>.<item>
    """

    # If its 'hidden', put it next last
    prefix = "z{}" if definition.name.startswith("_") else "a{}"
    return prefix.format(definition.name)



# Copyright 2021- Python Language Server Contributors.

import logging

from pylsp import _utils, hookimpl

log = logging.getLogger(__name__)


def lsp_location(name):
    module_path = name.module_path
    if module_path is None or name.line is None or name.column is None:
        return None
    uri = module_path.as_uri()
    return {
        "uri": str(uri),
        "range": {
            "start": {"line": name.line - 1, "character": name.column},
            "end": {"line": name.line - 1, "character": name.column + len(name.name)},
        },
    }


@hookimpl
def pylsp_type_definition(config, document, position):
    try:
        kwargs = _utils.position_to_jedi_linecolumn(document, position)
        script = document.jedi_script()
        names = script.infer(**kwargs)
        definitions = [
            definition
            for definition in [lsp_location(name) for name in names]
            if definition is not None
        ]
        return definitions
    except Exception as e:
        log.debug("Failed to run type_definition: %s", e)
        return []



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging

import pycodestyle
from autopep8 import continued_indentation as autopep8_c_i
from autopep8 import fix_code

from pylsp import hookimpl
from pylsp._utils import get_eol_chars

log = logging.getLogger(__name__)


@hookimpl(tryfirst=True)  # Prefer autopep8 over YAPF
def pylsp_format_document(config, workspace, document, options):
    with workspace.report_progress("format: autopep8"):
        log.info("Formatting document %s with autopep8", document)
        return _format(config, document)


@hookimpl(tryfirst=True)  # Prefer autopep8 over YAPF
def pylsp_format_range(config, workspace, document, range, options):
    log.info("Formatting document %s in range %s with autopep8", document, range)

    # First we 'round' the range up/down to full lines only
    range["start"]["character"] = 0
    range["end"]["line"] += 1
    range["end"]["character"] = 0

    # Add 1 for 1-indexing vs LSP's 0-indexing
    line_range = (range["start"]["line"] + 1, range["end"]["line"])
    return _format(config, document, line_range=line_range)


def _format(config, document, line_range=None):
    options = _autopep8_config(config, document)
    if line_range:
        options["line_range"] = list(line_range)

    # Temporarily re-monkey-patch the continued_indentation checker - #771
    del pycodestyle._checks["logical_line"][pycodestyle.continued_indentation]
    pycodestyle.register_check(autopep8_c_i)

    # Autopep8 doesn't work with CR line endings, so we replace them by '\n'
    # and restore them below.
    replace_cr = False
    source = document.source
    eol_chars = get_eol_chars(source)
    if eol_chars == "\r":
        replace_cr = True
        source = source.replace("\r", "\n")

    new_source = fix_code(source, options=options)

    # Switch it back
    del pycodestyle._checks["logical_line"][autopep8_c_i]
    pycodestyle.register_check(pycodestyle.continued_indentation)

    if new_source == source:
        return []

    if replace_cr:
        new_source = new_source.replace("\n", "\r")

    # I'm too lazy at the moment to parse diffs into TextEdit items
    # So let's just return the entire file...
    return [
        {
            "range": {
                "start": {"line": 0, "character": 0},
                # End char 0 of the line after our document
                "end": {"line": len(document.lines), "character": 0},
            },
            "newText": new_source,
        }
    ]


def _autopep8_config(config, document=None):
    # We user pycodestyle settings to avoid redefining things
    path = document.path if document is not None else None
    settings = config.plugin_settings("pycodestyle", document_path=path)
    options = {
        "exclude": settings.get("exclude"),
        "hang_closing": settings.get("hangClosing"),
        "ignore": settings.get("ignore"),
        "max_line_length": settings.get("maxLineLength"),
        "select": settings.get("select"),
        "aggressive": settings.get("aggressive"),
    }

    # Filter out null options
    return {k: v for k, v in options.items() if v}



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.


def get_well_formatted_range(lsp_range):
    start = lsp_range["start"]
    end = lsp_range["end"]

    if start["line"] > end["line"] or (
        start["line"] == end["line"] and start["character"] > end["character"]
    ):
        return {"start": end, "end": start}

    return lsp_range


def get_well_formatted_edit(text_edit):
    lsp_range = get_well_formatted_range(text_edit["range"])
    if lsp_range != text_edit["range"]:
        return {"newText": text_edit["newText"], "range": lsp_range}

    return text_edit


def compare_text_edits(a, b):
    diff = a["range"]["start"]["line"] - b["range"]["start"]["line"]
    if diff == 0:
        return a["range"]["start"]["character"] - b["range"]["start"]["character"]

    return diff


def merge_sort_text_edits(text_edits):
    if len(text_edits) <= 1:
        return text_edits

    p = len(text_edits) // 2
    left = text_edits[:p]
    right = text_edits[p:]

    merge_sort_text_edits(left)
    merge_sort_text_edits(right)

    left_idx = 0
    right_idx = 0
    i = 0
    while left_idx < len(left) and right_idx < len(right):
        ret = compare_text_edits(left[left_idx], right[right_idx])
        if ret <= 0:
            #  smaller_equal -> take left to preserve order
            text_edits[i] = left[left_idx]
            i += 1
            left_idx += 1
        else:
            # greater -> take right
            text_edits[i] = right[right_idx]
            i += 1
            right_idx += 1
    while left_idx < len(left):
        text_edits[i] = left[left_idx]
        i += 1
        left_idx += 1
    while right_idx < len(right):
        text_edits[i] = right[right_idx]
        i += 1
        right_idx += 1
    return text_edits


class OverLappingTextEditException(Exception):
    """
    Text edits are expected to be sorted
    and compressed instead of overlapping.
    This error is raised when two edits
    are overlapping.
    """


def apply_text_edits(doc, text_edits):
    text = doc.source
    sorted_edits = merge_sort_text_edits(list(map(get_well_formatted_edit, text_edits)))
    last_modified_offset = 0
    spans = []
    for e in sorted_edits:
        start_offset = doc.offset_at_position(e["range"]["start"])
        if start_offset < last_modified_offset:
            raise OverLappingTextEditException("overlapping edit")

        if start_offset > last_modified_offset:
            spans.append(text[last_modified_offset:start_offset])

        if len(e["newText"]):
            spans.append(e["newText"])
        last_modified_offset = doc.offset_at_position(e["range"]["end"])

    spans.append(text[last_modified_offset:])
    return "".join(spans)



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging
import os
import socketserver
import sys
import threading
import uuid
from functools import partial
from typing import Any

try:
    import ujson as json
except Exception:
    import json

from pylsp_jsonrpc.dispatchers import MethodDispatcher
from pylsp_jsonrpc.endpoint import Endpoint
from pylsp_jsonrpc.streams import JsonRpcStreamReader, JsonRpcStreamWriter

from . import _utils, lsp, uris
from ._version import __version__
from .config import config
from .workspace import Cell, Document, Notebook, Workspace

log = logging.getLogger(__name__)


LINT_DEBOUNCE_S = 0.5  # 500 ms
PARENT_PROCESS_WATCH_INTERVAL = 10  # 10 s
MAX_WORKERS = 64
PYTHON_FILE_EXTENSIONS = (".py", ".pyi")
CONFIG_FILEs = ("pycodestyle.cfg", "setup.cfg", "tox.ini", ".flake8")


class _StreamHandlerWrapper(socketserver.StreamRequestHandler):
    """A wrapper class that is used to construct a custom handler class."""

    delegate = None

    def setup(self) -> None:
        super().setup()
        self.delegate = self.DELEGATE_CLASS(self.rfile, self.wfile)

    def handle(self) -> None:
        try:
            self.delegate.start()
        except OSError as e:
            if os.name == "nt":
                # Catch and pass on ConnectionResetError when parent process
                # dies
                if isinstance(e, WindowsError) and e.winerror == 10054:
                    pass

        self.SHUTDOWN_CALL()


def start_tcp_lang_server(bind_addr, port, check_parent_process, handler_class) -> None:
    if not issubclass(handler_class, PythonLSPServer):
        raise ValueError("Handler class must be an instance of PythonLSPServer")

    def shutdown_server(check_parent_process, *args):
        if check_parent_process:
            log.debug("Shutting down server")
            # Shutdown call must be done on a thread, to prevent deadlocks
            stop_thread = threading.Thread(target=server.shutdown)
            stop_thread.start()

    # Construct a custom wrapper class around the user's handler_class
    wrapper_class = type(
        handler_class.__name__ + "Handler",
        (_StreamHandlerWrapper,),
        {
            # We need to wrap this in staticmethod due to the changes to
            # functools.partial in Python 3.14+
            "DELEGATE_CLASS": staticmethod(
                partial(handler_class, check_parent_process=check_parent_process)
            )
            if sys.version_info >= (3, 14)
            else partial(handler_class, check_parent_process=check_parent_process),
            "SHUTDOWN_CALL": partial(shutdown_server, check_parent_process),
        },
    )

    server = socketserver.TCPServer(
        (bind_addr, port), wrapper_class, bind_and_activate=False
    )
    server.allow_reuse_address = True

    try:
        server.server_bind()
        server.server_activate()
        log.info("Serving %s on (%s, %s)", handler_class.__name__, bind_addr, port)
        server.serve_forever()
    finally:
        log.info("Shutting down")
        server.server_close()


def start_io_lang_server(rfile, wfile, check_parent_process, handler_class) -> None:
    if not issubclass(handler_class, PythonLSPServer):
        raise ValueError("Handler class must be an instance of PythonLSPServer")
    log.info("Starting %s IO language server", handler_class.__name__)
    server = handler_class(rfile, wfile, check_parent_process)
    server.start()


def start_ws_lang_server(port, check_parent_process, handler_class) -> None:
    if not issubclass(handler_class, PythonLSPServer):
        raise ValueError("Handler class must be an instance of PythonLSPServer")

    # imports needed only for websockets based server
    try:
        import asyncio
        from concurrent.futures import ThreadPoolExecutor

        import websockets
    except ImportError as e:
        raise ImportError(
            "websocket modules missing. Please run: pip install 'python-lsp-server[websockets]'"
        ) from e

    with ThreadPoolExecutor(max_workers=10) as tpool:
        send_queue = None
        loop = None

        async def pylsp_ws(websocket):
            log.debug("Creating LSP object")

            # creating a partial function and suppling the websocket connection
            response_handler = partial(send_message, websocket=websocket)

            # Not using default stream reader and writer.
            # Instead using a consumer based approach to handle processed requests
            pylsp_handler = handler_class(
                rx=None,
                tx=None,
                consumer=response_handler,
                check_parent_process=check_parent_process,
            )

            async for message in websocket:
                try:
                    log.debug("consuming payload and feeding it to LSP handler")
                    request = json.loads(message)
                    loop = asyncio.get_running_loop()
                    await loop.run_in_executor(tpool, pylsp_handler.consume, request)
                except Exception as e:
                    log.exception("Failed to process request %s, %s", message, str(e))

        def send_message(message, websocket):
            """Handler to send responses of  processed requests to respective web socket clients"""
            try:
                payload = json.dumps(message, ensure_ascii=False)
                loop.call_soon_threadsafe(send_queue.put_nowait, (payload, websocket))
            except Exception as e:
                log.exception("Failed to write message %s, %s", message, str(e))

        async def run_server():
            nonlocal send_queue, loop
            send_queue = asyncio.Queue()
            loop = asyncio.get_running_loop()

            async with websockets.serve(pylsp_ws, port=port):
                while 1:
                    # Wait until payload is available for sending
                    payload, websocket = await send_queue.get()
                    await websocket.send(payload)

        asyncio.run(run_server())


class PythonLSPServer(MethodDispatcher):
    """Implementation of the Microsoft VSCode Language Server Protocol
    https://github.com/Microsoft/language-server-protocol/blob/master/versions/protocol-1-x.md
    """

    def __init__(
        self, rx, tx, check_parent_process=False, consumer=None, *, endpoint_cls=None
    ) -> None:
        self.workspace = None
        self.config = None
        self.root_uri = None
        self.watching_thread = None
        self.workspaces = {}
        self.uri_workspace_mapper = {}

        self._check_parent_process = check_parent_process

        if rx is not None:
            self._jsonrpc_stream_reader = JsonRpcStreamReader(rx)
        else:
            self._jsonrpc_stream_reader = None

        if tx is not None:
            self._jsonrpc_stream_writer = JsonRpcStreamWriter(tx)
        else:
            self._jsonrpc_stream_writer = None

        endpoint_cls = endpoint_cls or Endpoint

        # if consumer is None, it is assumed that the default streams-based approach is being used
        if consumer is None:
            self._endpoint = endpoint_cls(
                self, self._jsonrpc_stream_writer.write, max_workers=MAX_WORKERS
            )
        else:
            self._endpoint = endpoint_cls(self, consumer, max_workers=MAX_WORKERS)

        self._dispatchers = []
        self._shutdown = False

    def start(self) -> None:
        """Entry point for the server."""
        self._jsonrpc_stream_reader.listen(self._endpoint.consume)

    def consume(self, message) -> None:
        """Entry point for consumer based server. Alternative to stream listeners."""
        # assuming message will be JSON
        self._endpoint.consume(message)

    def __getitem__(self, item):
        """Override getitem to fallback through multiple dispatchers."""
        if self._shutdown and item != "exit":
            # exit is the only allowed method during shutdown
            log.debug("Ignoring non-exit method during shutdown: %s", item)
            item = "invalid_request_after_shutdown"

        try:
            return super().__getitem__(item)
        except KeyError:
            # Fallback through extra dispatchers
            for dispatcher in self._dispatchers:
                try:
                    return dispatcher[item]
                except KeyError:
                    continue

        raise KeyError()

    def m_shutdown(self, **_kwargs) -> None:
        for workspace in self.workspaces.values():
            workspace.close()
        self._hook("pylsp_shutdown")
        self._shutdown = True

    def m_invalid_request_after_shutdown(self, **_kwargs):
        return {
            "error": {
                "code": lsp.ErrorCodes.InvalidRequest,
                "message": "Requests after shutdown are not valid",
            }
        }

    def m_exit(self, **_kwargs) -> None:
        self._endpoint.shutdown()
        if self._jsonrpc_stream_reader is not None:
            self._jsonrpc_stream_reader.close()
        if self._jsonrpc_stream_writer is not None:
            self._jsonrpc_stream_writer.close()

    def _match_uri_to_workspace(self, uri):
        workspace_uri = _utils.match_uri_to_workspace(uri, self.workspaces)
        return self.workspaces.get(workspace_uri, self.workspace)

    def _hook(self, hook_name, doc_uri=None, **kwargs):
        """Calls hook_name and returns a list of results from all registered handlers"""
        workspace = self._match_uri_to_workspace(doc_uri)
        doc = workspace.get_document(doc_uri) if doc_uri else None
        hook_handlers = self.config.plugin_manager.subset_hook_caller(
            hook_name, self.config.disabled_plugins
        )
        return hook_handlers(
            config=self.config, workspace=workspace, document=doc, **kwargs
        )

    def capabilities(self):
        server_capabilities = {
            "codeActionProvider": True,
            "codeLensProvider": {
                "resolveProvider": False,  # We may need to make this configurable
            },
            "completionProvider": {
                "resolveProvider": True,  # We could know everything ahead of time, but this takes time to transfer
                "triggerCharacters": ["."],
            },
            "documentFormattingProvider": True,
            "documentHighlightProvider": True,
            "documentRangeFormattingProvider": True,
            "documentSymbolProvider": True,
            "definitionProvider": True,
            "typeDefinitionProvider": True,
            "executeCommandProvider": {
                "commands": flatten(self._hook("pylsp_commands"))
            },
            "hoverProvider": True,
            "referencesProvider": True,
            "renameProvider": True,
            "foldingRangeProvider": True,
            "signatureHelpProvider": {"triggerCharacters": ["(", ",", "="]},
            "textDocumentSync": {
                "change": lsp.TextDocumentSyncKind.INCREMENTAL,
                "save": {
                    "includeText": True,
                },
                "openClose": True,
            },
            "notebookDocumentSync": {
                "notebookSelector": [{"cells": [{"language": "python"}]}]
            },
            "workspace": {
                "workspaceFolders": {"supported": True, "changeNotifications": True}
            },
            "experimental": merge(self._hook("pylsp_experimental_capabilities")),
        }
        log.info("Server capabilities: %s", server_capabilities)
        return server_capabilities

    def m_initialize(
        self,
        processId=None,
        rootUri=None,
        rootPath=None,
        initializationOptions=None,
        workspaceFolders=None,
        **_kwargs,
    ):
        log.debug(
            "Language server initialized with %s %s %s %s",
            processId,
            rootUri,
            rootPath,
            initializationOptions,
        )
        if rootUri is None:
            rootUri = uris.from_fs_path(rootPath) if rootPath is not None else ""

        self.workspaces.pop(self.root_uri, None)
        self.root_uri = rootUri
        self.config = config.Config(
            rootUri,
            initializationOptions or {},
            processId,
            _kwargs.get("capabilities", {}),
        )
        self.workspace = Workspace(rootUri, self._endpoint, self.config)
        self.workspaces[rootUri] = self.workspace
        if workspaceFolders:
            for folder in workspaceFolders:
                uri = folder["uri"]
                if uri == rootUri:
                    # Already created
                    continue
                workspace_config = config.Config(
                    uri,
                    self.config._init_opts,
                    self.config._process_id,
                    self.config._capabilities,
                )
                workspace_config.update(self.config._settings)
                self.workspaces[uri] = Workspace(uri, self._endpoint, workspace_config)

        self._dispatchers = self._hook("pylsp_dispatchers")
        self._hook("pylsp_initialize")

        if (
            self._check_parent_process
            and processId is not None
            and self.watching_thread is None
        ):

            def watch_parent_process(pid):
                # exit when the given pid is not alive
                if not _utils.is_process_alive(pid):
                    log.info("parent process %s is not alive, exiting!", pid)
                    self.m_exit()
                else:
                    threading.Timer(
                        PARENT_PROCESS_WATCH_INTERVAL, watch_parent_process, args=[pid]
                    ).start()

            self.watching_thread = threading.Thread(
                target=watch_parent_process, args=(processId,)
            )
            self.watching_thread.daemon = True
            self.watching_thread.start()
        # Get our capabilities
        return {
            "capabilities": self.capabilities(),
            "serverInfo": {
                "name": "pylsp",
                "version": __version__,
            },
        }

    def m_initialized(self, **_kwargs) -> None:
        self._hook("pylsp_initialized")

    def code_actions(self, doc_uri: str, range: dict, context: dict):
        return flatten(
            self._hook("pylsp_code_actions", doc_uri, range=range, context=context)
        )

    def code_lens(self, doc_uri):
        return flatten(self._hook("pylsp_code_lens", doc_uri))

    def completions(self, doc_uri, position):
        workspace = self._match_uri_to_workspace(doc_uri)
        document = workspace.get_document(doc_uri)
        ignored_names = None
        if isinstance(document, Cell):
            # We need to get the ignored names from the whole notebook document
            notebook_document = workspace.get_maybe_document(document.notebook_uri)
            ignored_names = notebook_document.jedi_names(doc_uri)
        completions = self._hook(
            "pylsp_completions", doc_uri, position=position, ignored_names=ignored_names
        )
        return {"isIncomplete": False, "items": flatten(completions)}

    def completion_item_resolve(self, completion_item):
        doc_uri = completion_item.get("data", {}).get("doc_uri", None)
        return self._hook(
            "pylsp_completion_item_resolve", doc_uri, completion_item=completion_item
        )

    def definitions(self, doc_uri, position):
        return flatten(self._hook("pylsp_definitions", doc_uri, position=position))

    def type_definition(self, doc_uri, position):
        return self._hook("pylsp_type_definition", doc_uri, position=position)

    def document_symbols(self, doc_uri):
        return flatten(self._hook("pylsp_document_symbols", doc_uri))

    def document_did_save(self, doc_uri):
        return self._hook("pylsp_document_did_save", doc_uri)

    def execute_command(self, command, arguments):
        return self._hook("pylsp_execute_command", command=command, arguments=arguments)

    def format_document(self, doc_uri, options):
        return lambda: self._hook("pylsp_format_document", doc_uri, options=options)

    def format_range(self, doc_uri, range, options):
        return self._hook("pylsp_format_range", doc_uri, range=range, options=options)

    def highlight(self, doc_uri, position):
        return (
            flatten(self._hook("pylsp_document_highlight", doc_uri, position=position))
            or None
        )

    def hover(self, doc_uri, position):
        return self._hook("pylsp_hover", doc_uri, position=position) or {"contents": ""}

    @_utils.debounce(LINT_DEBOUNCE_S, keyed_by="doc_uri")
    def lint(self, doc_uri, is_saved) -> None:
        # Since we're debounced, the document may no longer be open
        workspace = self._match_uri_to_workspace(doc_uri)
        document_object = workspace.documents.get(doc_uri, None)
        if isinstance(document_object, Document):
            self._lint_text_document(
                doc_uri, workspace, is_saved, document_object.version
            )
        elif isinstance(document_object, Notebook):
            self._lint_notebook_document(document_object, workspace)

    def _lint_text_document(
        self, doc_uri, workspace, is_saved, doc_version=None
    ) -> None:
        workspace.publish_diagnostics(
            doc_uri,
            flatten(self._hook("pylsp_lint", doc_uri, is_saved=is_saved)),
            doc_version,
        )

    def _lint_notebook_document(self, notebook_document, workspace) -> None:
        """
        Lint a notebook document.

        This is a bit more complicated than linting a text document, because we need to
        send the entire notebook document to the pylsp_lint hook, but we need to send
        the diagnostics back to the client on a per-cell basis.
        """

        # First, we create a temp TextDocument that represents the whole notebook
        # contents. We'll use this to send to the pylsp_lint hook.
        random_uri = str(uuid.uuid4())

        # cell_list helps us map the diagnostics back to the correct cell later.
        cell_list: list[dict[str, Any]] = []

        offset = 0
        total_source = ""
        for cell in notebook_document.cells:
            cell_uri = cell["document"]
            cell_document = workspace.get_cell_document(cell_uri)

            num_lines = cell_document.line_count

            data = {
                "uri": cell_uri,
                "line_start": offset,
                "line_end": offset + num_lines - 1,
                "source": cell_document.source,
            }

            cell_list.append(data)
            if offset == 0:
                total_source = cell_document.source
            else:
                total_source += "\n" + cell_document.source

            offset += num_lines

        workspace.put_document(random_uri, total_source)

        try:
            document_diagnostics = flatten(
                self._hook("pylsp_lint", random_uri, is_saved=True)
            )

            # Now we need to map the diagnostics back to the correct cell and publish them.
            # Note: this is O(n*m) in the number of cells and diagnostics, respectively.
            for cell in cell_list:
                cell_diagnostics = []
                for diagnostic in document_diagnostics:
                    start_line = diagnostic["range"]["start"]["line"]
                    end_line = diagnostic["range"]["end"]["line"]

                    if start_line > cell["line_end"] or end_line < cell["line_start"]:
                        continue
                    diagnostic["range"]["start"]["line"] = (
                        start_line - cell["line_start"]
                    )
                    diagnostic["range"]["end"]["line"] = end_line - cell["line_start"]
                    cell_diagnostics.append(diagnostic)

                workspace.publish_diagnostics(cell["uri"], cell_diagnostics)
        finally:
            workspace.rm_document(random_uri)

    def references(self, doc_uri, position, exclude_declaration):
        return flatten(
            self._hook(
                "pylsp_references",
                doc_uri,
                position=position,
                exclude_declaration=exclude_declaration,
            )
        )

    def rename(self, doc_uri, position, new_name):
        return self._hook("pylsp_rename", doc_uri, position=position, new_name=new_name)

    def signature_help(self, doc_uri, position):
        return self._hook("pylsp_signature_help", doc_uri, position=position)

    def folding(self, doc_uri):
        return flatten(self._hook("pylsp_folding_range", doc_uri))

    def m_completion_item__resolve(self, **completionItem):
        return self.completion_item_resolve(completionItem)

    def m_notebook_document__did_open(
        self, notebookDocument=None, cellTextDocuments=None, **_kwargs
    ) -> None:
        workspace = self._match_uri_to_workspace(notebookDocument["uri"])
        workspace.put_notebook_document(
            notebookDocument["uri"],
            notebookDocument["notebookType"],
            cells=notebookDocument["cells"],
            version=notebookDocument.get("version"),
            metadata=notebookDocument.get("metadata"),
        )
        for cell in cellTextDocuments or []:
            workspace.put_cell_document(
                cell["uri"],
                notebookDocument["uri"],
                cell["languageId"],
                cell["text"],
                version=cell.get("version"),
            )
        self.lint(notebookDocument["uri"], is_saved=True)

    def m_notebook_document__did_close(
        self, notebookDocument=None, cellTextDocuments=None, **_kwargs
    ) -> None:
        workspace = self._match_uri_to_workspace(notebookDocument["uri"])
        for cell in cellTextDocuments or []:
            workspace.publish_diagnostics(cell["uri"], [])
            workspace.rm_document(cell["uri"])
        workspace.rm_document(notebookDocument["uri"])

    def m_notebook_document__did_change(
        self, notebookDocument=None, change=None, **_kwargs
    ) -> None:
        """
        Changes to the notebook document.

        This could be one of the following:
        1. Notebook metadata changed
        2. Cell(s) added
        3. Cell(s) deleted
        4. Cell(s) data changed
            4.1 Cell metadata changed
            4.2 Cell source changed
        """
        workspace = self._match_uri_to_workspace(notebookDocument["uri"])

        if change.get("metadata"):
            # Case 1
            workspace.update_notebook_metadata(
                notebookDocument["uri"], change.get("metadata")
            )

        cells = change.get("cells")
        if cells:
            # Change to cells
            structure = cells.get("structure")
            if structure:
                # Case 2 or 3
                notebook_cell_array_change = structure["array"]
                start = notebook_cell_array_change["start"]
                cell_delete_count = notebook_cell_array_change["deleteCount"]
                if cell_delete_count == 0:
                    # Case 2
                    # Cell documents
                    for cell_document in structure["didOpen"]:
                        workspace.put_cell_document(
                            cell_document["uri"],
                            notebookDocument["uri"],
                            cell_document["languageId"],
                            cell_document["text"],
                            cell_document.get("version"),
                        )
                    # Cell metadata which is added to Notebook
                    workspace.add_notebook_cells(
                        notebookDocument["uri"],
                        notebook_cell_array_change["cells"],
                        start,
                    )
                else:
                    # Case 3
                    # Cell documents
                    for cell_document in structure["didClose"]:
                        workspace.rm_document(cell_document["uri"])
                        workspace.publish_diagnostics(cell_document["uri"], [])
                    # Cell metadata which is removed from Notebook
                    workspace.remove_notebook_cells(
                        notebookDocument["uri"], start, cell_delete_count
                    )

            data = cells.get("data")
            if data:
                # Case 4.1
                for cell in data:
                    # update NotebookDocument.cells properties
                    pass

            text_content = cells.get("textContent")
            if text_content:
                # Case 4.2
                for cell in text_content:
                    cell_uri = cell["document"]["uri"]
                    # Even though the protocol says that `changes` is an array, we assume that it's always a single
                    # element array that contains the last change to the cell source.
                    workspace.update_document(cell_uri, cell["changes"][0])
        self.lint(notebookDocument["uri"], is_saved=True)

    def m_text_document__did_close(self, textDocument=None, **_kwargs) -> None:
        workspace = self._match_uri_to_workspace(textDocument["uri"])
        workspace.publish_diagnostics(textDocument["uri"], [])
        workspace.rm_document(textDocument["uri"])

    def m_text_document__did_open(self, textDocument=None, **_kwargs) -> None:
        workspace = self._match_uri_to_workspace(textDocument["uri"])
        workspace.put_document(
            textDocument["uri"],
            textDocument["text"],
            version=textDocument.get("version"),
        )
        self._hook("pylsp_document_did_open", textDocument["uri"])
        self.lint(textDocument["uri"], is_saved=True)

    def m_text_document__did_change(
        self, contentChanges=None, textDocument=None, **_kwargs
    ) -> None:
        workspace = self._match_uri_to_workspace(textDocument["uri"])
        for change in contentChanges:
            workspace.update_document(
                textDocument["uri"], change, version=textDocument.get("version")
            )
        self.lint(textDocument["uri"], is_saved=False)

    def m_text_document__did_save(self, textDocument=None, **_kwargs) -> None:
        self.lint(textDocument["uri"], is_saved=True)
        self.document_did_save(textDocument["uri"])

    def m_text_document__code_action(
        self, textDocument=None, range=None, context=None, **_kwargs
    ):
        return self.code_actions(textDocument["uri"], range, context)

    def m_text_document__code_lens(self, textDocument=None, **_kwargs):
        return self.code_lens(textDocument["uri"])

    def _cell_document__completion(self, cellDocument, position=None, **_kwargs):
        workspace = self._match_uri_to_workspace(cellDocument.notebook_uri)
        notebookDocument = workspace.get_maybe_document(cellDocument.notebook_uri)
        if notebookDocument is None:
            raise ValueError("Invalid notebook document")

        cell_data = notebookDocument.cell_data()

        # Concatenate all cells to be a single temporary document
        total_source = "\n".join(data["source"] for data in cell_data.values())
        with workspace.temp_document(total_source) as temp_uri:
            # update position to be the position in the temp document
            if position is not None:
                position["line"] += cell_data[cellDocument.uri]["line_start"]

            completions = self.completions(temp_uri, position)

            # Translate temp_uri locations to cell document locations
            for item in completions.get("items", []):
                if item.get("data", {}).get("doc_uri") == temp_uri:
                    item["data"]["doc_uri"] = cellDocument.uri

            # Copy LAST_JEDI_COMPLETIONS to cell document so that completionItem/resolve will work
            tempDocument = workspace.get_document(temp_uri)
            cellDocument.shared_data["LAST_JEDI_COMPLETIONS"] = (
                tempDocument.shared_data.get("LAST_JEDI_COMPLETIONS", None)
            )

            return completions

    def m_text_document__completion(self, textDocument=None, position=None, **_kwargs):
        # textDocument here is just a dict with a uri
        workspace = self._match_uri_to_workspace(textDocument["uri"])
        document = workspace.get_document(textDocument["uri"])
        if isinstance(document, Cell):
            return self._cell_document__completion(document, position, **_kwargs)
        return self.completions(textDocument["uri"], position)

    def _cell_document__definition(self, cellDocument, position=None, **_kwargs):
        workspace = self._match_uri_to_workspace(cellDocument.notebook_uri)
        notebookDocument = workspace.get_maybe_document(cellDocument.notebook_uri)
        if notebookDocument is None:
            raise ValueError("Invalid notebook document")

        cell_data = notebookDocument.cell_data()

        # Concatenate all cells to be a single temporary document
        total_source = "\n".join(data["source"] for data in cell_data.values())
        with workspace.temp_document(total_source) as temp_uri:
            # update position to be the position in the temp document
            if position is not None:
                position["line"] += cell_data[cellDocument.uri]["line_start"]

            definitions = self.definitions(temp_uri, position)

            # Translate temp_uri locations to cell document locations
            for definition in definitions:
                if definition["uri"] == temp_uri:
                    # Find the cell the start line is in and adjust the uri and line numbers
                    for cell_uri, data in cell_data.items():
                        if (
                            data["line_start"]
                            <= definition["range"]["start"]["line"]
                            <= data["line_end"]
                        ):
                            definition["uri"] = cell_uri
                            definition["range"]["start"]["line"] -= data["line_start"]
                            definition["range"]["end"]["line"] -= data["line_start"]
                            break

            return definitions

    def m_text_document__definition(self, textDocument=None, position=None, **_kwargs):
        # textDocument here is just a dict with a uri
        workspace = self._match_uri_to_workspace(textDocument["uri"])
        document = workspace.get_document(textDocument["uri"])
        if isinstance(document, Cell):
            return self._cell_document__definition(document, position, **_kwargs)
        return self.definitions(textDocument["uri"], position)

    def m_text_document__type_definition(
        self, textDocument=None, position=None, **_kwargs
    ):
        return self.type_definition(textDocument["uri"], position)

    def m_text_document__document_highlight(
        self, textDocument=None, position=None, **_kwargs
    ):
        return self.highlight(textDocument["uri"], position)

    def m_text_document__hover(self, textDocument=None, position=None, **_kwargs):
        return self.hover(textDocument["uri"], position)

    def m_text_document__document_symbol(self, textDocument=None, **_kwargs):
        return self.document_symbols(textDocument["uri"])

    def m_text_document__formatting(self, textDocument=None, options=None, **_kwargs):
        return self.format_document(textDocument["uri"], options)

    def m_text_document__rename(
        self, textDocument=None, position=None, newName=None, **_kwargs
    ):
        return self.rename(textDocument["uri"], position, newName)

    def m_text_document__folding_range(self, textDocument=None, **_kwargs):
        return self.folding(textDocument["uri"])

    def m_text_document__range_formatting(
        self, textDocument=None, range=None, options=None, **_kwargs
    ):
        return self.format_range(textDocument["uri"], range, options)

    def m_text_document__references(
        self, textDocument=None, position=None, context=None, **_kwargs
    ):
        exclude_declaration = not context["includeDeclaration"]
        return self.references(textDocument["uri"], position, exclude_declaration)

    def m_text_document__signature_help(
        self, textDocument=None, position=None, **_kwargs
    ):
        return self.signature_help(textDocument["uri"], position)

    def m_workspace__did_change_configuration(self, settings=None) -> None:
        if self.config is not None:
            self.config.update((settings or {}).get("pylsp", {}))
        for workspace in self.workspaces.values():
            workspace.update_config(settings)
            self._hook("pylsp_workspace_configuration_changed")
            for doc_uri in workspace.documents:
                self.lint(doc_uri, is_saved=False)

    def m_workspace__did_change_workspace_folders(self, event=None, **_kwargs):
        if event is None:
            return
        added = event.get("added", [])
        removed = event.get("removed", [])

        for removed_info in removed:
            if "uri" in removed_info:
                removed_uri = removed_info["uri"]
                self.workspaces.pop(removed_uri, None)

        for added_info in added:
            if "uri" in added_info:
                added_uri = added_info["uri"]
                workspace_config = config.Config(
                    added_uri,
                    self.config._init_opts,
                    self.config._process_id,
                    self.config._capabilities,
                )
                workspace_config.update(self.config._settings)
                self.workspaces[added_uri] = Workspace(
                    added_uri, self._endpoint, workspace_config
                )

        root_workspace_removed = any(
            removed_info["uri"] == self.root_uri for removed_info in removed
        )
        workspace_added = len(added) > 0 and "uri" in added[0]
        if root_workspace_removed and workspace_added:
            added_uri = added[0]["uri"]
            self.root_uri = added_uri
            new_root_workspace = self.workspaces[added_uri]
            self.config = new_root_workspace._config
            self.workspace = new_root_workspace
        elif root_workspace_removed:
            # NOTE: Removing the root workspace can only happen when the server
            # is closed, thus the else condition of this if can never happen.
            if self.workspaces:
                log.debug("Root workspace deleted!")
                available_workspaces = sorted(self.workspaces)
                first_workspace = available_workspaces[0]
                new_root_workspace = self.workspaces[first_workspace]
                self.root_uri = first_workspace
                self.config = new_root_workspace._config
                self.workspace = new_root_workspace

        # Migrate documents that are on the root workspace and have a better
        # match now
        doc_uris = list(self.workspace._docs.keys())
        for uri in doc_uris:
            doc = self.workspace._docs.pop(uri)
            new_workspace = self._match_uri_to_workspace(uri)
            new_workspace._docs[uri] = doc

    def m_workspace__did_change_watched_files(self, changes=None, **_kwargs):
        changed_py_files = set()
        config_changed = False
        for d in changes or []:
            if d["uri"].endswith(PYTHON_FILE_EXTENSIONS):
                changed_py_files.add(d["uri"])
            elif d["uri"].endswith(CONFIG_FILEs):
                config_changed = True

        if config_changed:
            self.config.settings.cache_clear()
        elif not changed_py_files:
            # Only externally changed python files and lint configs may result in changed diagnostics.
            return

        for workspace in self.workspaces.values():
            for doc_uri in workspace.documents:
                # Changes in doc_uri are already handled by m_text_document__did_save
                if doc_uri not in changed_py_files:
                    self.lint(doc_uri, is_saved=False)

    def m_workspace__execute_command(self, command=None, arguments=None):
        return self.execute_command(command, arguments)


def flatten(list_of_lists):
    return [item for lst in list_of_lists for item in lst]


def merge(list_of_dicts):
    return {k: v for dictionary in list_of_dicts for k, v in dictionary.items()}



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging

from pylsp import _utils, hookimpl, uris

log = logging.getLogger(__name__)


@hookimpl
def pylsp_rename(config, workspace, document, position, new_name):
    log.debug(
        "Executing rename of %s to %s", document.word_at_position(position), new_name
    )
    kwargs = _utils.position_to_jedi_linecolumn(document, position)
    kwargs["new_name"] = new_name
    try:
        refactoring = document.jedi_script().rename(**kwargs)
    except NotImplementedError as exc:
        raise Exception(
            "No support for renaming in Python 2/3.5 with Jedi. "
            "Consider using the pylsp-rope plugin instead"
        ) from exc
    log.debug("Finished rename: %s", refactoring.get_diff())
    changes = []

    changed_files = refactoring.get_changed_files()
    for file_path, changed_file in changed_files.items():
        uri = uris.from_fs_path(str(file_path))
        doc = workspace.get_maybe_document(uri)
        changes.append(
            {
                "textDocument": {"uri": uri, "version": doc.version if doc else None},
                "edits": [
                    {
                        "range": {
                            "start": {"line": 0, "character": 0},
                            "end": {
                                "line": _num_lines(changed_file.get_new_code()),
                                "character": 0,
                            },
                        },
                        "newText": changed_file.get_new_code(),
                    }
                ],
            }
        )
    return {"documentChanges": changes}


def _num_lines(file_contents):
    "Count the number of lines in the given string."
    if _utils.get_eol_chars(file_contents):
        return len(file_contents.splitlines())
    return 0



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

from pyflakes import api as pyflakes_api
from pyflakes import messages

from pylsp import hookimpl, lsp

# Pyflakes messages that should be reported as Errors instead of Warns
PYFLAKES_ERROR_MESSAGES = (
    messages.UndefinedName,
    messages.UndefinedExport,
    messages.UndefinedLocal,
    messages.DuplicateArgument,
    messages.FutureFeatureNotDefined,
    messages.ReturnOutsideFunction,
    messages.YieldOutsideFunction,
    messages.ContinueOutsideLoop,
    messages.BreakOutsideLoop,
    messages.TwoStarredExpressions,
)


@hookimpl
def pylsp_lint(workspace, document):
    with workspace.report_progress("lint: pyflakes"):
        reporter = PyflakesDiagnosticReport(document.lines)
        pyflakes_api.check(
            document.source.encode("utf-8"), document.path, reporter=reporter
        )
        return reporter.diagnostics


class PyflakesDiagnosticReport:
    def __init__(self, lines) -> None:
        self.lines = lines
        self.diagnostics = []

    def unexpectedError(self, _filename, msg) -> None:  # pragma: no cover
        err_range = {
            "start": {"line": 0, "character": 0},
            "end": {"line": 0, "character": 0},
        }
        self.diagnostics.append(
            {
                "source": "pyflakes",
                "range": err_range,
                "message": msg,
                "severity": lsp.DiagnosticSeverity.Error,
            }
        )

    def syntaxError(self, _filename, msg, lineno, offset, text) -> None:
        # We've seen that lineno and offset can sometimes be None
        lineno = lineno or 1
        offset = offset or 0
        # could be None if the error is due to an invalid encoding
        # see e.g. https://github.com/python-lsp/python-lsp-server/issues/429
        text = text or ""

        err_range = {
            "start": {"line": lineno - 1, "character": offset},
            "end": {"line": lineno - 1, "character": offset + len(text)},
        }
        self.diagnostics.append(
            {
                "source": "pyflakes",
                "range": err_range,
                "message": msg,
                "severity": lsp.DiagnosticSeverity.Error,
            }
        )

    def flake(self, message) -> None:
        """Get message like <filename>:<lineno>: <msg>"""
        err_range = {
            "start": {"line": message.lineno - 1, "character": message.col},
            "end": {
                "line": message.lineno - 1,
                "character": len(self.lines[message.lineno - 1]),
            },
        }

        severity = lsp.DiagnosticSeverity.Warning
        for message_type in PYFLAKES_ERROR_MESSAGES:
            if isinstance(message, message_type):
                severity = lsp.DiagnosticSeverity.Error
                break

        self.diagnostics.append(
            {
                "source": "pyflakes",
                "range": err_range,
                "message": message.message % message.message_args,
                "severity": severity,
            }
        )



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import configparser
import logging
import os
import sys

log = logging.getLogger(__name__)


class ConfigSource:
    """Base class for implementing a config source."""

    def __init__(self, root_path) -> None:
        self.root_path = root_path
        self.is_windows = sys.platform == "win32"
        self.xdg_home = os.environ.get(
            "XDG_CONFIG_HOME", os.path.expanduser("~/.config")
        )

    def user_config(self) -> None:
        """Return user-level (i.e. home directory) configuration."""
        raise NotImplementedError()

    def project_config(self, document_path) -> None:
        """Return project-level (i.e. workspace directory) configuration."""
        raise NotImplementedError()

    @classmethod
    def read_config_from_files(cls, files):
        config = configparser.RawConfigParser()
        for filename in files:
            if os.path.exists(filename) and not os.path.isdir(filename):
                config.read(filename)

        return config

    @classmethod
    def parse_config(cls, config, key, options):
        """Parse the config with the given options."""
        conf = {}
        for source, destination, opt_type in options:
            opt_value = cls._get_opt(config, key, source, opt_type)
            if opt_value is not None:
                cls._set_opt(conf, destination, opt_value)
        return conf

    @classmethod
    def _get_opt(cls, config, key, option, opt_type):
        """Get an option from a configparser with the given type."""
        for opt_key in [option, option.replace("-", "_")]:
            if not config.has_option(key, opt_key):
                continue

            if opt_type is bool:
                return config.getboolean(key, opt_key)

            if opt_type is int:
                return config.getint(key, opt_key)

            if opt_type is str:
                return config.get(key, opt_key)

            if opt_type is list:
                return cls._parse_list_opt(config.get(key, opt_key))

            raise ValueError("Unknown option type: %s" % opt_type)

    @classmethod
    def _parse_list_opt(cls, string):
        return [s.strip() for s in string.split(",") if s.strip()]

    @classmethod
    def _set_opt(cls, config_dict, path, value):
        """Set the value in the dictionary at the given path if the value is not None."""
        if value is None:
            return

        if "." not in path:
            config_dict[path] = value
            return

        key, rest = path.split(".", 1)
        if key not in config_dict:
            config_dict[key] = {}

        cls._set_opt(config_dict[key], rest, value)



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

"""Some Language Server Protocol constants

https://github.com/Microsoft/language-server-protocol/blob/master/protocol.md
"""


class CompletionItemKind:
    Text = 1
    Method = 2
    Function = 3
    Constructor = 4
    Field = 5
    Variable = 6
    Class = 7
    Interface = 8
    Module = 9
    Property = 10
    Unit = 11
    Value = 12
    Enum = 13
    Keyword = 14
    Snippet = 15
    Color = 16
    File = 17
    Reference = 18
    Folder = 19
    EnumMember = 20
    Constant = 21
    Struct = 22
    Event = 23
    Operator = 24
    TypeParameter = 25


class DocumentHighlightKind:
    Text = 1
    Read = 2
    Write = 3


class DiagnosticSeverity:
    Error = 1
    Warning = 2
    Information = 3
    Hint = 4


class DiagnosticTag:
    Unnecessary = 1
    Deprecated = 2


class InsertTextFormat:
    PlainText = 1
    Snippet = 2


class MessageType:
    Error = 1
    Warning = 2
    Info = 3
    Log = 4


class SymbolKind:
    File = 1
    Module = 2
    Namespace = 3
    Package = 4
    Class = 5
    Method = 6
    Property = 7
    Field = 8
    Constructor = 9
    Enum = 10
    Interface = 11
    Function = 12
    Variable = 13
    Constant = 14
    String = 15
    Number = 16
    Boolean = 17
    Array = 18


class TextDocumentSyncKind:
    NONE = 0
    FULL = 1
    INCREMENTAL = 2


class NotebookCellKind:
    Markup = 1
    Code = 2


# https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#errorCodes
class ErrorCodes:
    ParseError = -32700
    InvalidRequest = -32600
    MethodNotFound = -32601
    InvalidParams = -32602
    InternalError = -32603
    jsonrpcReservedErrorRangeStart = -32099
    ServerNotInitialized = -32002
    UnknownErrorCode = -32001
    jsonrpcReservedErrorRangeEnd = -32000
    lspReservedErrorRangeStart = -32899
    ServerCancelled = -32802
    ContentModified = -32801
    RequestCancelled = -32800
    lspReservedErrorRangeEnd = -32800



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import contextlib
import logging
import os
import re
import sys

import pydocstyle

from pylsp import hookimpl, lsp

log = logging.getLogger(__name__)

# PyDocstyle is a little verbose in debug message
pydocstyle_logger = logging.getLogger(pydocstyle.utils.__name__)
pydocstyle_logger.setLevel(logging.INFO)

DEFAULT_MATCH_RE = pydocstyle.config.ConfigurationParser.DEFAULT_MATCH_RE
DEFAULT_MATCH_DIR_RE = pydocstyle.config.ConfigurationParser.DEFAULT_MATCH_DIR_RE


@hookimpl
def pylsp_settings():
    # Default pydocstyle to disabled
    return {"plugins": {"pydocstyle": {"enabled": False}}}


@hookimpl
def pylsp_lint(config, workspace, document):
    with workspace.report_progress("lint: pydocstyle"):
        settings = config.plugin_settings("pydocstyle", document_path=document.path)
        log.debug("Got pydocstyle settings: %s", settings)

        # Explicitly passing a path to pydocstyle means it doesn't respect the --match flag, so do it ourselves
        filename_match_re = re.compile(settings.get("match", DEFAULT_MATCH_RE) + "$")
        if not filename_match_re.match(os.path.basename(document.path)):
            return []

        # Likewise with --match-dir
        dir_match_re = re.compile(settings.get("matchDir", DEFAULT_MATCH_DIR_RE) + "$")
        if not dir_match_re.match(os.path.basename(os.path.dirname(document.path))):
            return []

        args = [document.path]

        if settings.get("convention"):
            args.append("--convention=" + settings["convention"])

            if settings.get("addSelect"):
                args.append("--add-select=" + ",".join(settings["addSelect"]))
            if settings.get("addIgnore"):
                args.append("--add-ignore=" + ",".join(settings["addIgnore"]))

        elif settings.get("select"):
            args.append("--select=" + ",".join(settings["select"]))
        elif settings.get("ignore"):
            args.append("--ignore=" + ",".join(settings["ignore"]))

        log.info("Using pydocstyle args: %s", args)

        conf = pydocstyle.config.ConfigurationParser()
        with _patch_sys_argv(args):
            # TODO(gatesn): We can add more pydocstyle args here from our pylsp config
            conf.parse()

        # Will only yield a single filename, the document path
        diags = []
        for (
            filename,
            checked_codes,
            ignore_decorators,
            property_decorators,
            ignore_self_only_init,
        ) in conf.get_files_to_check():
            errors = pydocstyle.checker.ConventionChecker().check_source(
                document.source,
                filename,
                ignore_decorators=ignore_decorators,
                property_decorators=property_decorators,
                ignore_self_only_init=ignore_self_only_init,
            )

            try:
                for error in errors:
                    if error.code not in checked_codes:
                        continue
                    diags.append(_parse_diagnostic(document, error))
            except pydocstyle.parser.ParseError:
                # In the case we cannot parse the Python file, just continue
                pass

        log.debug("Got pydocstyle errors: %s", diags)
        return diags


def _parse_diagnostic(document, error):
    lineno = error.definition.start - 1
    line = document.lines[0] if document.lines else ""

    start_character = len(line) - len(line.lstrip())
    end_character = len(line)

    return {
        "source": "pydocstyle",
        "code": error.code,
        "message": error.message,
        "severity": lsp.DiagnosticSeverity.Warning,
        "range": {
            "start": {"line": lineno, "character": start_character},
            "end": {"line": lineno, "character": end_character},
        },
    }


@contextlib.contextmanager
def _patch_sys_argv(arguments) -> None:
    old_args = sys.argv

    # Preserve argv[0] since it's the executable
    sys.argv = old_args[0:1] + arguments

    try:
        yield
    finally:
        sys.argv = old_args



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging

from pylsp import _utils, hookimpl

log = logging.getLogger(__name__)


@hookimpl
def pylsp_hover(config, document, position):
    signature_config = config.settings().get("signature", {})
    code_position = _utils.position_to_jedi_linecolumn(document, position)
    definitions = document.jedi_script(use_document_path=True).infer(**code_position)
    word = document.word_at_position(position)

    # Find first exact matching definition
    definition = next((x for x in definitions if x.name == word), None)

    # Ensure a definition is used if only one is available
    # even if the word doesn't match. An example of this case is 'np'
    # where 'numpy' doesn't match with 'np'. Same for NumPy ufuncs
    if len(definitions) == 1:
        definition = definitions[0]

    if not definition:
        return {"contents": ""}

    hover_capabilities = config.capabilities.get("textDocument", {}).get("hover", {})
    supported_markup_kinds = hover_capabilities.get("contentFormat", ["markdown"])
    preferred_markup_kind = _utils.choose_markup_kind(supported_markup_kinds)

    # Find first exact matching signature
    signature = next(
        (
            x.to_string()
            for x in definition.get_signatures()
            if (x.name == word and x.type not in ["module"])
        ),
        "",
    )

    include_docstring = signature_config.get("include_docstring", True)

    # raw docstring returns only doc, without signature
    docstring = definition.docstring(raw=True)
    if not include_docstring:
        if signature:
            docstring = ""
        else:
            docstring = docstring.strip().split("\n")[0].strip()

    return {
        "contents": _utils.format_docstring(
            docstring,
            preferred_markup_kind,
            signatures=[signature] if signature else None,
            signature_config=signature_config,
        )
    }



# Copyright 2018 Google LLC.
# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

"""Linter plugin for pylint."""

import collections
import logging
import os
import re
import shlex
import sys
from subprocess import PIPE, Popen

from pylsp import hookimpl, lsp

try:
    import ujson as json
except Exception:
    import json

log = logging.getLogger(__name__)

# Pylint fails to suppress STDOUT when importing whitelisted C
# extensions, mangling their output into the expected JSON which breaks the
# parser. The most prominent example (and maybe the only one out there) is
# pygame - we work around that by asking pygame to NOT display the message upon
# import via an (otherwise harmless) environment variable. This is an ad-hoc
# fix for a very specific upstream issue.
# Related: https://github.com/PyCQA/pylint/issues/3518
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = "hide"
DEPRECATION_CODES = {
    "W0402",  # Uses of a deprecated module %r
    "W1505",  # Using deprecated method %s()
    "W1511",  # Using deprecated argument %s of method %s()
    "W1512",  # Using deprecated class %s of module %s
    "W1513",  # Using deprecated decorator %s()
}
UNNECESSITY_CODES = {
    "W0611",  # Unused import %s
    "W0612",  # Unused variable %r
    "W0613",  # Unused argument %r
    "W0614",  # Unused import %s from wildcard import
    "W1304",  # Unused-format-string-argument
}


class PylintLinter:
    last_diags = collections.defaultdict(list)

    @classmethod
    def lint(cls, document, is_saved, flags=""):
        """Plugin interface to pylsp linter.

        Args:
            document: The document to be linted.
            is_saved: Whether or not the file has been saved to disk.
            flags: Additional flags to pass to pylint. Not exposed to
                pylsp_lint, but used for testing.

        Returns:
            A list of dicts with the following format:

                {
                    'source': 'pylint',
                    'range': {
                        'start': {
                            'line': start_line,
                            'character': start_column,
                        },
                        'end': {
                            'line': end_line,
                            'character': end_column,
                        },
                    }
                    'message': msg,
                    'severity': lsp.DiagnosticSeverity.*,
                }
        """
        if not is_saved:
            # Pylint can only be run on files that have been saved to disk.
            # Rather than return nothing, return the previous list of
            # diagnostics. If we return an empty list, any diagnostics we'd
            # previously shown will be cleared until the next save. Instead,
            # continue showing (possibly stale) diagnostics until the next
            # save.
            return cls.last_diags[document.path]

        cmd = [
            sys.executable,
            "-c",
            "import sys; from pylint.lint import Run; Run(sys.argv[1:])",
            "-f",
            "json",
            document.path,
        ] + (shlex.split(str(flags)) if flags else [])
        log.debug("Calling pylint with '%s'", " ".join(cmd))

        cwd = document._workspace.root_path
        if not cwd:
            cwd = os.path.dirname(__file__)

        with Popen(
            cmd, stdout=PIPE, stderr=PIPE, cwd=cwd, universal_newlines=True
        ) as process:
            json_out, err = process.communicate()

        if err != "":
            log.error("Error calling pylint: '%s'", err)

        # pylint prints nothing rather than [] when there are no diagnostics.
        # json.loads will not parse an empty string, so just return.
        if not json_out.strip():
            cls.last_diags[document.path] = []
            return []

        # Pylint's JSON output is a list of objects with the following format.
        #
        #     {
        #         "obj": "main",
        #         "path": "foo.py",
        #         "message": "Missing function docstring",
        #         "message-id": "C0111",
        #         "symbol": "missing-docstring",
        #         "column": 0,
        #         "type": "convention",
        #         "line": 5,
        #         "module": "foo"
        #     }
        #
        # The type can be any of:
        #
        #  * convention
        #  * information
        #  * error
        #  * fatal
        #  * refactor
        #  * warning
        diagnostics = []
        for diag in json.loads(json_out):
            # pylint lines index from 1, pylsp lines index from 0
            line = diag["line"] - 1

            err_range = {
                "start": {
                    "line": line,
                    # Index columns start from 0
                    "character": diag["column"],
                },
                "end": {
                    "line": line,
                    # It's possible that we're linting an empty file. Even an empty
                    # file might fail linting if it isn't named properly.
                    "character": (
                        _find_end_of_identifier(document.lines[line], diag["column"])
                        if document.lines
                        else 0
                    ),
                },
            }

            if diag["type"] == "convention":
                severity = lsp.DiagnosticSeverity.Information
            elif diag["type"] == "information":
                severity = lsp.DiagnosticSeverity.Information
            elif diag["type"] == "error":
                severity = lsp.DiagnosticSeverity.Error
            elif diag["type"] == "fatal":
                severity = lsp.DiagnosticSeverity.Error
            elif diag["type"] == "refactor":
                severity = lsp.DiagnosticSeverity.Hint
            elif diag["type"] == "warning":
                severity = lsp.DiagnosticSeverity.Warning

            code = diag["message-id"]

            diagnostic = {
                "source": "pylint",
                "range": err_range,
                "message": "[{}] {}".format(diag["symbol"], diag["message"]),
                "severity": severity,
                "code": code,
            }

            if code in UNNECESSITY_CODES:
                diagnostic["tags"] = [lsp.DiagnosticTag.Unnecessary]
            if code in DEPRECATION_CODES:
                diagnostic["tags"] = [lsp.DiagnosticTag.Deprecated]

            diagnostics.append(diagnostic)
        cls.last_diags[document.path] = diagnostics
        return diagnostics


def _build_pylint_flags(settings):
    """Build arguments for calling pylint."""
    pylint_args = settings.get("args")
    if pylint_args is None:
        return ""
    return " ".join(pylint_args)


@hookimpl
def pylsp_settings():
    # Default pylint to disabled because it requires a config
    # file to be useful.
    return {
        "plugins": {
            "pylint": {
                "enabled": False,
                "args": [],
                # disabled by default as it can slow down the workflow
                "executable": None,
            }
        }
    }


@hookimpl
def pylsp_lint(config, workspace, document, is_saved):
    """Run pylint linter."""
    with workspace.report_progress("lint: pylint"):
        settings = config.plugin_settings("pylint")
        log.debug("Got pylint settings: %s", settings)
        # pylint >= 2.5.0 is required for working through stdin and only
        # available with python3
        if settings.get("executable") and sys.version_info[0] >= 3:
            flags = build_args_stdio(settings)
            pylint_executable = settings.get("executable", "pylint")
            return pylint_lint_stdin(pylint_executable, document, flags)
        flags = _build_pylint_flags(settings)
        return PylintLinter.lint(document, is_saved, flags=flags)


def build_args_stdio(settings):
    """Build arguments for calling pylint.

    :param settings: client settings
    :type settings: dict

    :return: arguments to path to pylint
    :rtype: list
    """
    pylint_args = settings.get("args")
    if pylint_args is None:
        return []
    return pylint_args


def pylint_lint_stdin(pylint_executable, document, flags):
    """Run pylint linter from stdin.

    This runs pylint in a subprocess with popen.
    This allows passing the file from stdin and as a result
    run pylint on unsaved files. Can slowdown the workflow.

    :param pylint_executable: path to pylint executable
    :type pylint_executable: string
    :param document: document to run pylint on
    :type document: pylsp.workspace.Document
    :param flags: arguments to path to pylint
    :type flags: list

    :return: linting diagnostics
    :rtype: list
    """
    pylint_result = _run_pylint_stdio(pylint_executable, document, flags)
    return _parse_pylint_stdio_result(document, pylint_result)


def _run_pylint_stdio(pylint_executable, document, flags):
    """Run pylint in popen.

    :param pylint_executable: path to pylint executable
    :type pylint_executable: string
    :param document: document to run pylint on
    :type document: pylsp.workspace.Document
    :param flags: arguments to path to pylint
    :type flags: list

    :return: result of calling pylint
    :rtype: string
    """
    log.debug("Calling %s with args: '%s'", pylint_executable, flags)
    try:
        cmd = [pylint_executable]
        cmd.extend(flags)
        cmd.extend(["--from-stdin", document.path])
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    except OSError:
        log.debug("Can't execute %s. Trying with 'python -m pylint'", pylint_executable)
        cmd = [sys.executable, "-m", "pylint"]
        cmd.extend(flags)
        cmd.extend(["--from-stdin", document.path])
        p = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate(document.source.encode())
    if stderr:
        log.error("Error while running pylint '%s'", stderr.decode())
    return stdout.decode()


def _parse_pylint_stdio_result(document, stdout):
    """Parse pylint results.

    :param document: document to run pylint on
    :type document: pylsp.workspace.Document
    :param stdout: pylint results to parse
    :type stdout: string

    :return: linting diagnostics
    :rtype: list
    """
    diagnostics = []
    lines = stdout.splitlines()
    for raw_line in lines:
        parsed_line = re.match(r"(.*):(\d*):(\d*): (\w*): (.*)", raw_line)
        if not parsed_line:
            log.debug("Pylint output parser can't parse line '%s'", raw_line)
            continue

        parsed_line = parsed_line.groups()
        if len(parsed_line) != 5:
            log.debug("Pylint output parser can't parse line '%s'", raw_line)
            continue

        _, line, character, code, msg = parsed_line
        line = int(line) - 1
        character = int(character)
        severity_map = {
            "C": lsp.DiagnosticSeverity.Information,
            "E": lsp.DiagnosticSeverity.Error,
            "F": lsp.DiagnosticSeverity.Error,
            "I": lsp.DiagnosticSeverity.Information,
            "R": lsp.DiagnosticSeverity.Hint,
            "W": lsp.DiagnosticSeverity.Warning,
        }
        severity = severity_map[code[0]]
        diagnostic = {
            "source": "pylint",
            "code": code,
            "range": {
                "start": {"line": line, "character": character},
                "end": {
                    "line": line,
                    "character": _find_end_of_identifier(
                        document.lines[line], character
                    ),
                },
            },
            "message": msg,
            "severity": severity,
        }
        if code in UNNECESSITY_CODES:
            diagnostic["tags"] = [lsp.DiagnosticTag.Unnecessary]
        if code in DEPRECATION_CODES:
            diagnostic["tags"] = [lsp.DiagnosticTag.Deprecated]
        diagnostics.append(diagnostic)

    return diagnostics


def _find_end_of_identifier(string, start):
    """Find the end of the identifier starting at the given position."""
    for i in range(len(string), start, -1):
        if string[start:i].isidentifier():
            return i
    return len(string) - 1



# Copyright 2022- Python Language Server Contributors.

import logging
import threading
from collections.abc import Generator
from typing import Any, Optional, Union

import parso
from jedi import Script
from parso.python import tree
from parso.tree import NodeOrLeaf
from rope.base.resources import Resource
from rope.contrib.autoimport.defs import SearchResult
from rope.contrib.autoimport.sqlite import AutoImport

from pylsp import hookimpl
from pylsp.config.config import Config
from pylsp.workspace import Document, Workspace

from ._rope_task_handle import PylspTaskHandle

log = logging.getLogger(__name__)

_score_pow = 5
_score_max = 10**_score_pow
MAX_RESULTS_COMPLETIONS = 1000
MAX_RESULTS_CODE_ACTIONS = 5


class AutoimportCache:
    """Handles the cache creation."""

    def __init__(self) -> None:
        self.thread = None

    def reload_cache(
        self,
        config: Config,
        workspace: Workspace,
        files: Optional[list[Document]] = None,
        single_thread: Optional[bool] = True,
    ):
        if self.is_blocked():
            return

        memory: bool = config.plugin_settings("rope_autoimport").get("memory", False)
        rope_config = config.settings().get("rope", {})
        autoimport = workspace._rope_autoimport(rope_config, memory)
        resources: Optional[list[Resource]] = (
            None
            if files is None
            else [document._rope_resource(rope_config) for document in files]
        )

        if single_thread:
            self._reload_cache(workspace, autoimport, resources)
        else:
            # Creating the cache may take 10-20s for a environment with 5k python modules. That's
            # why we decided to move cache creation into its own thread.
            self.thread = threading.Thread(
                target=self._reload_cache, args=(workspace, autoimport, resources)
            )
            self.thread.start()

    def _reload_cache(
        self,
        workspace: Workspace,
        autoimport: AutoImport,
        resources: Optional[list[Resource]] = None,
    ) -> None:
        task_handle = PylspTaskHandle(workspace)
        autoimport.generate_cache(task_handle=task_handle, resources=resources)
        autoimport.generate_modules_cache(task_handle=task_handle)

    def is_blocked(self):
        return self.thread and self.thread.is_alive()


@hookimpl
def pylsp_settings() -> dict[str, dict[str, dict[str, Any]]]:
    # Default rope_completion to disabled
    return {
        "plugins": {
            "rope_autoimport": {
                "enabled": False,
                "memory": False,
                "completions": {
                    "enabled": True,
                },
                "code_actions": {
                    "enabled": True,
                },
            }
        }
    }


def _should_insert(expr: tree.BaseNode, word_node: tree.Leaf) -> bool:
    """
    Check if we should insert the word_node on the given expr.

    Works for both correct and incorrect code. This is because the
    user is often working on the code as they write it.
    """
    if not word_node:
        return False
    if len(expr.children) == 0:
        return True
    first_child = expr.children[0]
    if isinstance(first_child, tree.EndMarker):
        if "#" in first_child.prefix:
            return False  # Check for single line comment
    if first_child == word_node:
        return True  # If the word is the first word then its fine
    if len(expr.children) > 1:
        if any(
            node.type == "operator" and "." in node.value or node.type == "trailer"
            for node in expr.children
        ):
            return False  # Check if we're on a method of a function
    if isinstance(first_child, (tree.PythonErrorNode, tree.PythonNode)):
        # The tree will often include error nodes like this to indicate errors
        # we want to ignore errors since the code is being written
        return _should_insert(first_child, word_node)
    return _handle_first_child(first_child, expr, word_node)


def _handle_first_child(
    first_child: NodeOrLeaf, expr: tree.BaseNode, word_node: tree.Leaf
) -> bool:
    """Check if we suggest imports given the following first child."""
    if isinstance(first_child, tree.Import):
        return False
    if isinstance(first_child, (tree.PythonLeaf, tree.PythonErrorLeaf)):
        # Check if the first item is a from or import statement even when incomplete
        if first_child.value in ("import", "from"):
            return False
    if isinstance(first_child, tree.Keyword):
        if first_child.value == "def":
            return _should_import_function(word_node, expr)
        if first_child.value == "class":
            return _should_import_class(word_node, expr)
    return True


def _should_import_class(word_node: tree.Leaf, expr: tree.BaseNode) -> bool:
    prev_node = None
    for node in expr.children:
        if isinstance(node, tree.Name):
            if isinstance(prev_node, tree.Operator):
                if node == word_node and prev_node.value == "(":
                    return True
        prev_node = node

    return False


def _should_import_function(word_node: tree.Leaf, expr: tree.BaseNode) -> bool:
    prev_node = None
    for node in expr.children:
        if _handle_argument(node, word_node):
            return True
        if isinstance(prev_node, tree.Operator):
            if prev_node.value == "->":
                if node == word_node:
                    return True
        prev_node = node
    return False


def _handle_argument(node: NodeOrLeaf, word_node: tree.Leaf):
    if isinstance(node, tree.PythonNode):
        if node.type == "tfpdef":
            if node.children[2] == word_node:
                return True
        if node.type == "parameters":
            for parameter in node.children:
                if _handle_argument(parameter, word_node):
                    return True
    return False


def _process_statements(
    suggestions: list[SearchResult],
    doc_uri: str,
    word: str,
    autoimport: AutoImport,
    document: Document,
    feature: str = "completions",
) -> Generator[dict[str, Any], None, None]:
    for suggestion in suggestions:
        insert_line = autoimport.find_insertion_line(document.source) - 1
        start = {"line": insert_line, "character": 0}
        edit_range = {"start": start, "end": start}
        edit = {"range": edit_range, "newText": suggestion.import_statement + "\n"}
        score = _get_score(
            suggestion.source, suggestion.import_statement, suggestion.name, word
        )
        if score > _score_max:
            continue
        # TODO make this markdown
        if feature == "completions":
            yield {
                "label": suggestion.name,
                "kind": suggestion.itemkind,
                "sortText": _sort_import(score),
                "data": {"doc_uri": doc_uri},
                "detail": _document(suggestion.import_statement),
                "additionalTextEdits": [edit],
            }
        elif feature == "code_actions":
            yield {
                "title": suggestion.import_statement,
                "kind": "quickfix",
                "edit": {"changes": {doc_uri: [edit]}},
                # data is a supported field for codeAction responses
                # See https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#textDocument_codeAction
                "data": {"sortText": _sort_import(score)},
            }
        else:
            raise ValueError(f"Unknown feature: {feature}")


def get_names(script: Script) -> set[str]:
    """Get all names to ignore from the current file."""
    raw_names = script.get_names(definitions=True)
    log.debug(raw_names)
    return {name.name for name in raw_names}


@hookimpl
def pylsp_completions(
    config: Config,
    workspace: Workspace,
    document: Document,
    position,
    ignored_names: Union[set[str], None],
):
    """Get autoimport suggestions."""
    if (
        not config.plugin_settings("rope_autoimport")
        .get("completions", {})
        .get("enabled", True)
    ) or cache.is_blocked():
        return []

    line = document.lines[position["line"]]
    expr = parso.parse(line)
    word_node = expr.get_leaf_for_position((1, position["character"]))
    if not _should_insert(expr, word_node):
        return []
    word = word_node.value
    log.debug(f"autoimport: searching for word: {word}")
    rope_config = config.settings(document_path=document.path).get("rope", {})
    ignored_names: set[str] = ignored_names or get_names(
        document.jedi_script(use_document_path=True)
    )
    autoimport = workspace._rope_autoimport(rope_config)
    suggestions = list(autoimport.search_full(word, ignored_names=ignored_names))
    results = sorted(
        _process_statements(
            suggestions, document.uri, word, autoimport, document, "completions"
        ),
        key=lambda statement: statement["sortText"],
    )
    if len(results) > MAX_RESULTS_COMPLETIONS:
        results = results[:MAX_RESULTS_COMPLETIONS]
    return results


def _document(import_statement: str) -> str:
    return """# Auto-Import\n""" + import_statement


def _get_score(
    source: int, full_statement: str, suggested_name: str, desired_name
) -> int:
    import_length = len("import")
    full_statement_score = len(full_statement) - import_length
    suggested_name_score = (len(suggested_name) - len(desired_name)) ** 2
    source_score = 20 * source
    return suggested_name_score + full_statement_score + source_score


def _sort_import(score: int) -> str:
    score = max(min(score, (_score_max) - 1), 0)
    # Since we are using ints, we need to pad them.
    # We also want to prioritize autoimport behind everything since its the last priority.
    # The minimum is to prevent score from overflowing the pad
    return "[z" + str(score).rjust(_score_pow, "0")


def get_name_or_module(document, diagnostic) -> str:
    start = diagnostic["range"]["start"]
    return (
        parso.parse(document.lines[start["line"]])
        .get_leaf_for_position((1, start["character"] + 1))
        .value
    )


@hookimpl
def pylsp_code_actions(
    config: Config,
    workspace: Workspace,
    document: Document,
    range: dict,
    context: dict,
) -> list[dict]:
    """
    Provide code actions through rope.

    Parameters
    ----------
    config : pylsp.config.config.Config
        Current config.
    workspace : pylsp.workspace.Workspace
        Current workspace.
    document : pylsp.workspace.Document
        Document to apply code actions on.
    range : dict
        Range argument given by pylsp. Not used here.
    context : dict
        CodeActionContext given as dict.

    Returns
    -------
      List of dicts containing the code actions.
    """
    if (
        not config.plugin_settings("rope_autoimport")
        .get("code_actions", {})
        .get("enabled", True)
    ) or cache.is_blocked():
        return []

    log.debug(f"textDocument/codeAction: {document} {range} {context}")
    code_actions = []
    for diagnostic in context.get("diagnostics", []):
        if "undefined name" not in diagnostic.get("message", "").lower():
            continue

        word = get_name_or_module(document, diagnostic)
        log.debug(f"autoimport: searching for word: {word}")
        rope_config = config.settings(document_path=document.path).get("rope", {})
        autoimport = workspace._rope_autoimport(rope_config)
        suggestions = list(autoimport.search_full(word))
        log.debug("autoimport: suggestions: %s", suggestions)
        results = sorted(
            _process_statements(
                suggestions,
                document.uri,
                word,
                autoimport,
                document,
                "code_actions",
            ),
            key=lambda statement: statement["data"]["sortText"],
        )

        if len(results) > MAX_RESULTS_CODE_ACTIONS:
            results = results[:MAX_RESULTS_CODE_ACTIONS]
        code_actions.extend(results)

    return code_actions


@hookimpl
def pylsp_initialize(config: Config, workspace: Workspace) -> None:
    """Initialize AutoImport.

    Generates the cache for local and global items.
    """
    cache.reload_cache(config, workspace)


@hookimpl
def pylsp_document_did_open(config: Config, workspace: Workspace) -> None:
    """Initialize AutoImport.

    Generates the cache for local and global items.
    """
    cache.reload_cache(config, workspace)


@hookimpl
def pylsp_document_did_save(
    config: Config, workspace: Workspace, document: Document
) -> None:
    """Update the names associated with this document."""
    cache.reload_cache(config, workspace, [document])


@hookimpl
def pylsp_workspace_configuration_changed(config: Config, workspace: Workspace) -> None:
    """
    Initialize autoimport if it has been enabled through a
    workspace/didChangeConfiguration message from the frontend.

    Generates the cache for local and global items.
    """
    if config.plugin_settings("rope_autoimport").get("enabled", False):
        cache.reload_cache(config, workspace)
    else:
        log.debug("autoimport: Skipping cache reload.")


cache: AutoimportCache = AutoimportCache()



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import pycodestyle

from pylsp._utils import find_parents

from .source import ConfigSource

CONFIG_KEY = "pycodestyle"
USER_CONFIGS = [pycodestyle.USER_CONFIG] if pycodestyle.USER_CONFIG else []
PROJECT_CONFIGS = ["pycodestyle.cfg", "setup.cfg", "tox.ini"]

OPTIONS = [
    ("exclude", "plugins.pycodestyle.exclude", list),
    ("filename", "plugins.pycodestyle.filename", list),
    ("hang-closing", "plugins.pycodestyle.hangClosing", bool),
    ("ignore", "plugins.pycodestyle.ignore", list),
    ("max-line-length", "plugins.pycodestyle.maxLineLength", int),
    ("indent-size", "plugins.pycodestyle.indentSize", int),
    ("select", "plugins.pycodestyle.select", list),
    ("aggressive", "plugins.pycodestyle.aggressive", int),
]


class PyCodeStyleConfig(ConfigSource):
    def user_config(self):
        config = self.read_config_from_files(USER_CONFIGS)
        return self.parse_config(config, CONFIG_KEY, OPTIONS)

    def project_config(self, document_path):
        files = find_parents(self.root_path, document_path, PROJECT_CONFIGS)
        config = self.read_config_from_files(files)
        return self.parse_config(config, CONFIG_KEY, OPTIONS)



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import functools
import io
import logging
import os
import re
import uuid
from collections.abc import Generator
from contextlib import contextmanager
from threading import RLock
from typing import Callable, Optional

import jedi

from . import _utils, lsp, uris

log = logging.getLogger(__name__)

DEFAULT_AUTO_IMPORT_MODULES = ["numpy"]

# TODO: this is not the best e.g. we capture numbers
RE_START_WORD = re.compile("[A-Za-z_0-9]*$")
RE_END_WORD = re.compile("^[A-Za-z_0-9]*")


def lock(method):
    """Define an atomic region over a method."""

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        with self._lock:
            return method(self, *args, **kwargs)

    return wrapper


class Workspace:
    M_PUBLISH_DIAGNOSTICS = "textDocument/publishDiagnostics"
    M_PROGRESS = "$/progress"
    M_INITIALIZE_PROGRESS = "window/workDoneProgress/create"
    M_APPLY_EDIT = "workspace/applyEdit"
    M_SHOW_MESSAGE = "window/showMessage"
    M_LOG_MESSAGE = "window/logMessage"

    def __init__(self, root_uri, endpoint, config=None) -> None:
        self._config = config
        self._root_uri = root_uri
        self._endpoint = endpoint
        self._root_uri_scheme = uris.urlparse(self._root_uri)[0]
        self._root_path = uris.to_fs_path(self._root_uri)
        self._docs = {}

        # Cache jedi environments
        self._environments = {}

        # Whilst incubating, keep rope private
        self.__rope = None
        self.__rope_config = None
        self.__rope_autoimport = None

    def _rope_autoimport(
        self,
        rope_config: Optional,
        memory: bool = False,
    ):
        from rope.contrib.autoimport.sqlite import AutoImport

        if self.__rope_autoimport is None:
            project = self._rope_project_builder(rope_config)
            self.__rope_autoimport = AutoImport(project, memory=memory)
        return self.__rope_autoimport

    def _rope_project_builder(self, rope_config):
        from rope.base.project import Project

        # TODO: we could keep track of dirty files and validate only those
        if self.__rope is None or self.__rope_config != rope_config:
            rope_folder = rope_config.get("ropeFolder")
            if "ropeFolder" in rope_config:
                self.__rope = Project(self._root_path, ropefolder=rope_folder)
            else:
                self.__rope = Project(self._root_path)
            self.__rope.prefs.set(
                "extension_modules", rope_config.get("extensionModules", [])
            )
            self.__rope.prefs.set("ignore_syntax_errors", True)
            self.__rope.prefs.set("ignore_bad_imports", True)
        self.__rope.validate()
        return self.__rope

    @property
    def documents(self):
        return self._docs

    @property
    def root_path(self):
        return self._root_path

    @property
    def root_uri(self):
        return self._root_uri

    def is_local(self):
        return (self._root_uri_scheme in ["", "file"]) and os.path.exists(
            self._root_path
        )

    def get_document(self, doc_uri):
        """Return a managed document if-present, else create one pointing at disk.

        See https://github.com/Microsoft/language-server-protocol/issues/177
        """
        return self._docs.get(doc_uri) or self._create_document(doc_uri)

    def get_cell_document(self, doc_uri):
        return self._docs.get(doc_uri)

    def get_maybe_document(self, doc_uri):
        return self._docs.get(doc_uri)

    def put_document(self, doc_uri, source, version=None) -> None:
        self._docs[doc_uri] = self._create_document(
            doc_uri, source=source, version=version
        )

    def put_notebook_document(
        self, doc_uri, notebook_type, cells, version=None, metadata=None
    ) -> None:
        self._docs[doc_uri] = self._create_notebook_document(
            doc_uri, notebook_type, cells, version, metadata
        )

    @contextmanager
    def temp_document(self, source, path=None) -> None:
        if path is None:
            path = self.root_path
        uri = uris.from_fs_path(os.path.join(path, str(uuid.uuid4())))
        try:
            self.put_document(uri, source)
            yield uri
        finally:
            self.rm_document(uri)

    def add_notebook_cells(self, doc_uri, cells, start) -> None:
        self._docs[doc_uri].add_cells(cells, start)

    def remove_notebook_cells(self, doc_uri, start, delete_count) -> None:
        self._docs[doc_uri].remove_cells(start, delete_count)

    def update_notebook_metadata(self, doc_uri, metadata) -> None:
        self._docs[doc_uri].metadata = metadata

    def put_cell_document(
        self, doc_uri, notebook_uri, language_id, source, version=None
    ) -> None:
        self._docs[doc_uri] = self._create_cell_document(
            doc_uri, notebook_uri, language_id, source, version
        )

    def rm_document(self, doc_uri) -> None:
        self._docs.pop(doc_uri)

    def update_document(self, doc_uri, change, version=None) -> None:
        self._docs[doc_uri].apply_change(change)
        self._docs[doc_uri].version = version

    def update_config(self, settings):
        self._config.update((settings or {}).get("pylsp", {}))
        for doc_uri in self.documents:
            if isinstance(document := self.get_document(doc_uri), Notebook):
                # Notebook documents don't have a config. The config is
                # handled at the cell level.
                return
            document.update_config(settings)

    def apply_edit(self, edit):
        return self._endpoint.request(self.M_APPLY_EDIT, {"edit": edit})

    def publish_diagnostics(self, doc_uri, diagnostics, doc_version=None) -> None:
        params = {
            "uri": doc_uri,
            "diagnostics": diagnostics,
        }

        if doc_version:
            params["version"] = doc_version

        self._endpoint.notify(
            self.M_PUBLISH_DIAGNOSTICS,
            params=params,
        )

    @contextmanager
    def report_progress(
        self,
        title: str,
        message: Optional[str] = None,
        percentage: Optional[int] = None,
        skip_token_initialization: bool = False,
    ) -> Generator[Callable[[str, Optional[int]], None], None, None]:
        """
        Report progress to the editor / client.

        ``skip_token_initialization` is necessary due to some current
        limitations of our LSP implementation. When `report_progress`
        is used from a synchronous LSP handler, the token initialization
        will time out because we can't receive the response.

        Many editors will still correctly show the progress messages though, which
        is why we are giving progress users the option to skip the initialization
        of the progress token.
        """
        if self._config:
            client_supports_progress_reporting = self._config.capabilities.get(
                "window", {}
            ).get("workDoneProgress", False)
        else:
            client_supports_progress_reporting = False

        if client_supports_progress_reporting:
            token = self._progress_begin(
                title, message, percentage, skip_token_initialization
            )

            def progress_message(
                message: str, percentage: Optional[int] = None
            ) -> None:
                self._progress_report(token, message, percentage)

            try:
                yield progress_message
            finally:
                self._progress_end(token)

            return

        # FALLBACK:
        # If the client doesn't support progress reporting, we have a dummy method
        # for the caller to use.
        def dummy_progress_message(
            message: str, percentage: Optional[int] = None
        ) -> None:
            pass

        yield dummy_progress_message

    def _progress_begin(
        self,
        title: str,
        message: Optional[str] = None,
        percentage: Optional[int] = None,
        skip_token_initialization: bool = False,
    ) -> str:
        token = str(uuid.uuid4())

        if not skip_token_initialization:
            try:
                self._endpoint.request(
                    self.M_INITIALIZE_PROGRESS, {"token": token}
                ).result(timeout=1.0)
            except Exception:
                log.warning(
                    "There was an error while trying to initialize progress reporting."
                    "Likely progress reporting was used in a synchronous LSP handler, "
                    "which is not supported by progress reporting yet. "
                    "To prevent waiting for the timeout you can set "
                    "`skip_token_initialization=True`. "
                    "Not every editor will show progress then, but many will.",
                    exc_info=True,
                )

        value = {
            "kind": "begin",
            "title": title,
        }
        if message is not None:
            value["message"] = message
        if percentage is not None:
            value["percentage"] = percentage

        self._endpoint.notify(
            self.M_PROGRESS,
            params={
                "token": token,
                "value": value,
            },
        )
        return token

    def _progress_report(
        self,
        token: str,
        message: Optional[str] = None,
        percentage: Optional[int] = None,
    ) -> None:
        value = {
            "kind": "report",
        }
        if message:
            value["message"] = message
        if percentage:
            value["percentage"] = percentage

        self._endpoint.notify(
            self.M_PROGRESS,
            params={
                "token": token,
                "value": value,
            },
        )

    def _progress_end(self, token: str, message: Optional[str] = None) -> None:
        value = {
            "kind": "end",
        }
        if message:
            value["message"] = message

        self._endpoint.notify(
            self.M_PROGRESS,
            params={
                "token": token,
                "value": value,
            },
        )

    def log_message(self, message, msg_type=lsp.MessageType.Info):
        self._endpoint.notify(
            self.M_LOG_MESSAGE, params={"type": msg_type, "message": message}
        )

    def show_message(self, message, msg_type=lsp.MessageType.Info) -> None:
        self._endpoint.notify(
            self.M_SHOW_MESSAGE, params={"type": msg_type, "message": message}
        )

    def source_roots(self, document_path):
        """Return the source roots for the given document."""
        files = (
            _utils.find_parents(
                self._root_path, document_path, ["setup.py", "pyproject.toml"]
            )
            or []
        )
        return list({os.path.dirname(project_file) for project_file in files}) or [
            self._root_path
        ]

    def _create_document(self, doc_uri, source=None, version=None):
        path = uris.to_fs_path(doc_uri)
        return Document(
            doc_uri,
            self,
            source=source,
            version=version,
            extra_sys_path=self.source_roots(path),
            rope_project_builder=self._rope_project_builder,
        )

    def _create_notebook_document(
        self, doc_uri, notebook_type, cells, version=None, metadata=None
    ):
        return Notebook(
            doc_uri,
            notebook_type,
            self,
            cells=cells,
            version=version,
            metadata=metadata,
        )

    def _create_cell_document(
        self, doc_uri, notebook_uri, language_id, source=None, version=None
    ):
        # TODO: remove what is unnecessary here.
        path = uris.to_fs_path(doc_uri)
        return Cell(
            doc_uri,
            notebook_uri=notebook_uri,
            language_id=language_id,
            workspace=self,
            source=source,
            version=version,
            extra_sys_path=self.source_roots(path),
            rope_project_builder=self._rope_project_builder,
        )

    def close(self) -> None:
        if self.__rope_autoimport:
            self.__rope_autoimport.close()


class Document:
    def __init__(
        self,
        uri,
        workspace,
        source=None,
        version=None,
        local=True,
        extra_sys_path=None,
        rope_project_builder=None,
    ) -> None:
        self.uri = uri
        self.version = version
        self.path = uris.to_fs_path(uri)
        self.dot_path = _utils.path_to_dot_name(self.path)
        self.filename = os.path.basename(self.path)
        self.shared_data = {}

        self._config = workspace._config
        self._workspace = workspace
        self._local = local
        self._source = source
        self._extra_sys_path = extra_sys_path or []
        self._rope_project_builder = rope_project_builder
        self._lock = RLock()

    def __str__(self):
        return str(self.uri)

    def _rope_resource(self, rope_config):
        from rope.base import libutils

        return libutils.path_to_resource(
            self._rope_project_builder(rope_config), self.path
        )

    @property
    @lock
    def lines(self):
        return self.source.splitlines(True)

    @property
    @lock
    def source(self):
        if self._source is None:
            with open(self.path, encoding="utf-8") as f:
                return f.read()
        return self._source

    def update_config(self, settings) -> None:
        self._config.update((settings or {}).get("pylsp", {}))

    @lock
    def apply_change(self, change):
        """Apply a change to the document."""
        text = change["text"]
        change_range = change.get("range")

        if not change_range:
            # The whole file has changed
            self._source = text
            return

        start_line = change_range["start"]["line"]
        start_col = change_range["start"]["character"]
        end_line = change_range["end"]["line"]
        end_col = change_range["end"]["character"]

        # Check for an edit occuring at the very end of the file
        lines = self.lines
        if start_line == len(lines):
            self._source = self.source + text
            return

        new = io.StringIO()

        # Iterate over the existing document until we hit the edit range,
        # at which point we write the new text, then loop until we hit
        # the end of the range and continue writing.
        for i, line in enumerate(lines):
            if i < start_line:
                new.write(line)
                continue

            if i > end_line:
                new.write(line)
                continue

            if i == start_line:
                new.write(line[:start_col])
                new.write(text)

            if i == end_line:
                new.write(line[end_col:])

        self._source = new.getvalue()

    def offset_at_position(self, position):
        """Return the byte-offset pointed at by the given position."""
        return position["character"] + len("".join(self.lines[: position["line"]]))

    def word_at_position(self, position):
        """Get the word under the cursor returning the start and end positions."""
        lines = self.lines
        if position["line"] >= len(lines):
            return ""

        line = lines[position["line"]]
        i = position["character"]
        # Split word in two
        start = line[:i]
        end = line[i:]

        # Take end of start and start of end to find word
        # These are guaranteed to match, even if they match the empty string
        m_start = RE_START_WORD.findall(start)
        m_end = RE_END_WORD.findall(end)

        return m_start[0] + m_end[-1]

    @lock
    def jedi_names(self, all_scopes=False, definitions=True, references=False):
        script = self.jedi_script()
        return script.get_names(
            all_scopes=all_scopes, definitions=definitions, references=references
        )

    @lock
    def jedi_script(self, position=None, use_document_path=False):
        extra_paths = []
        environment_path = None
        env_vars = None
        prioritize_extra_paths = False

        if self._config:
            jedi_settings = self._config.plugin_settings(
                "jedi", document_path=self.path
            )
            jedi.settings.auto_import_modules = jedi_settings.get(
                "auto_import_modules", DEFAULT_AUTO_IMPORT_MODULES
            )
            environment_path = jedi_settings.get("environment")
            # Jedi itself cannot deal with homedir-relative paths.
            # On systems, where it is expected, expand the home directory.
            if environment_path and os.name != "nt":
                environment_path = os.path.expanduser(environment_path)

            extra_paths = jedi_settings.get("extra_paths") or []
            env_vars = jedi_settings.get("env_vars")
            prioritize_extra_paths = jedi_settings.get("prioritize_extra_paths")

        # Drop PYTHONPATH from env_vars before creating the environment to
        # ensure that Jedi can startup properly without module name collision.
        if env_vars is None:
            env_vars = os.environ.copy()
        env_vars.pop("PYTHONPATH", None)

        environment = self.get_enviroment(environment_path, env_vars=env_vars)
        sys_path = self.sys_path(
            environment_path, env_vars, prioritize_extra_paths, extra_paths
        )

        project_path = self._workspace.root_path

        # Extend sys_path with document's path if requested
        if use_document_path:
            sys_path += [os.path.normpath(os.path.dirname(self.path))]

        kwargs = {
            "code": self.source,
            "path": self.path,
            "environment": environment if environment_path else None,
            "project": jedi.Project(path=project_path, sys_path=sys_path),
        }

        if position:
            # Deprecated by Jedi to use in Script() constructor
            kwargs += _utils.position_to_jedi_linecolumn(self, position)

        return jedi.Script(**kwargs)

    def get_enviroment(self, environment_path=None, env_vars=None):
        # TODO(gatesn): #339 - make better use of jedi environments, they seem pretty powerful
        if environment_path is None:
            environment = jedi.api.environment.get_cached_default_environment()
        else:
            if environment_path in self._workspace._environments:
                environment = self._workspace._environments[environment_path]
            else:
                environment = jedi.api.environment.create_environment(
                    path=environment_path, safe=False, env_vars=env_vars
                )
                self._workspace._environments[environment_path] = environment

        return environment

    def sys_path(
        self,
        environment_path=None,
        env_vars=None,
        prioritize_extra_paths=False,
        extra_paths=[],
    ):
        # Copy our extra sys path
        path = list(self._extra_sys_path)
        environment = self.get_enviroment(
            environment_path=environment_path, env_vars=env_vars
        )
        path.extend(environment.get_sys_path())
        if prioritize_extra_paths:
            path = extra_paths + path
        else:
            path = path + extra_paths

        return path


class Notebook:
    """Represents a notebook."""

    def __init__(
        self, uri, notebook_type, workspace, cells=None, version=None, metadata=None
    ) -> None:
        self.uri = uri
        self.notebook_type = notebook_type
        self.workspace = workspace
        self.version = version
        self.cells = cells or []
        self.metadata = metadata or {}
        self._lock = RLock()

    def __str__(self):
        return "Notebook with URI '%s'" % str(self.uri)

    def add_cells(self, new_cells: list, start: int) -> None:
        self.cells[start:start] = new_cells

    def remove_cells(self, start: int, delete_count: int) -> None:
        del self.cells[start : start + delete_count]

    def cell_data(self):
        """Extract current cell data.

        Returns a dict (ordered by cell position) where the key is the cell uri and the
        value is a dict with line_start, line_end, and source attributes.
        """
        cell_data = {}
        offset = 0
        for cell in self.cells:
            cell_uri = cell["document"]
            cell_document = self.workspace.get_cell_document(cell_uri)
            num_lines = cell_document.line_count
            cell_data[cell_uri] = {
                "line_start": offset,
                "line_end": offset + num_lines - 1,
                "source": cell_document.source,
            }
            offset += num_lines
        return cell_data

    @lock
    def jedi_names(
        self,
        up_to_cell_uri: Optional[str] = None,
        all_scopes=False,
        definitions=True,
        references=False,
    ):
        """
        Get the names in the notebook up to a certain cell.

        Parameters
        ----------
        up_to_cell_uri: str, optional
            The cell uri to stop at. If None, all cells are considered.
        """
        names = set()
        for cell in self.cells:
            cell_uri = cell["document"]
            cell_document = self.workspace.get_cell_document(cell_uri)
            names.update(cell_document.jedi_names(all_scopes, definitions, references))
            if cell_uri == up_to_cell_uri:
                break
        return {name.name for name in names}


class Cell(Document):
    """
    Represents a cell in a notebook.

    Notes
    -----
    We inherit from Document for now to get the same API. However, a cell document differs from text documents in that
    they have a language id.
    """

    def __init__(
        self,
        uri,
        notebook_uri,
        language_id,
        workspace,
        source=None,
        version=None,
        local=True,
        extra_sys_path=None,
        rope_project_builder=None,
    ) -> None:
        super().__init__(
            uri, workspace, source, version, local, extra_sys_path, rope_project_builder
        )
        self.language_id = language_id
        self.notebook_uri = notebook_uri

    @property
    @lock
    def line_count(self):
        """ "Return the number of lines in the cell document."""
        return len(self.source.split("\n"))



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging

from pylsp import _utils, hookimpl, uris

log = logging.getLogger(__name__)


@hookimpl
def pylsp_references(document, position, exclude_declaration):
    code_position = _utils.position_to_jedi_linecolumn(document, position)
    usages = document.jedi_script().get_references(**code_position)

    if exclude_declaration:
        # Filter out if the usage is the actual declaration of the thing
        usages = [d for d in usages if not d.is_definition()]

    # Filter out builtin modules
    return [
        {
            "uri": uris.uri_with(document.uri, path=str(d.module_path))
            if d.module_path
            else document.uri,
            "range": {
                "start": {"line": d.line - 1, "character": d.column},
                "end": {"line": d.line - 1, "character": d.column + len(d.name)},
            },
        }
        for d in usages
        if not d.in_builtin_module()
    ]



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging

from rope.contrib.codeassist import code_assist, sorted_proposals

from pylsp import _utils, hookimpl, lsp

log = logging.getLogger(__name__)


@hookimpl
def pylsp_settings():
    # Default rope_completion to disabled
    return {"plugins": {"rope_completion": {"enabled": False, "eager": False}}}


def _resolve_completion(completion, data, markup_kind):
    try:
        doc = _utils.format_docstring(data.get_doc(), markup_kind=markup_kind)
    except Exception as e:
        log.debug("Failed to resolve Rope completion: %s", e)
        doc = ""
    completion["detail"] = "{} {}".format(data.scope or "", data.name)
    completion["documentation"] = doc
    return completion


@hookimpl
def pylsp_completions(config, workspace, document, position):
    settings = config.plugin_settings("rope_completion", document_path=document.path)
    resolve_eagerly = settings.get("eager", False)

    # Rope is a bit rubbish at completing module imports, so we'll return None
    word = document.word_at_position(
        {
            # The -1 should really be trying to look at the previous word, but that might be quite expensive
            # So we only skip import completions when the cursor is one space after `import`
            "line": position["line"],
            "character": max(position["character"] - 1, 0),
        }
    )
    if word == "import":
        return None

    offset = document.offset_at_position(position)
    rope_config = config.settings(document_path=document.path).get("rope", {})
    rope_project = workspace._rope_project_builder(rope_config)
    document_rope = document._rope_resource(rope_config)

    completion_capabilities = config.capabilities.get("textDocument", {}).get(
        "completion", {}
    )
    item_capabilities = completion_capabilities.get("completionItem", {})
    supported_markup_kinds = item_capabilities.get("documentationFormat", ["markdown"])
    preferred_markup_kind = _utils.choose_markup_kind(supported_markup_kinds)

    try:
        definitions = code_assist(
            rope_project, document.source, offset, document_rope, maxfixes=3
        )
    except Exception as e:
        log.debug("Failed to run Rope code assist: %s", e)
        return []

    definitions = sorted_proposals(definitions)
    new_definitions = []
    for d in definitions:
        item = {
            "label": d.name,
            "kind": _kind(d),
            "sortText": _sort_text(d),
            "data": {"doc_uri": document.uri},
        }
        if resolve_eagerly:
            item = _resolve_completion(item, d, preferred_markup_kind)
        new_definitions.append(item)

    # most recently retrieved completion items, used for resolution
    document.shared_data["LAST_ROPE_COMPLETIONS"] = {
        # label is the only required property; here it is assumed to be unique
        completion["label"]: (completion, data)
        for completion, data in zip(new_definitions, definitions)
    }

    definitions = new_definitions

    return definitions or None


@hookimpl
def pylsp_completion_item_resolve(config, completion_item, document):
    """Resolve formatted completion for given non-resolved completion"""
    shared_data = document.shared_data["LAST_ROPE_COMPLETIONS"].get(
        completion_item["label"]
    )

    completion_capabilities = config.capabilities.get("textDocument", {}).get(
        "completion", {}
    )
    item_capabilities = completion_capabilities.get("completionItem", {})
    supported_markup_kinds = item_capabilities.get("documentationFormat", ["markdown"])
    preferred_markup_kind = _utils.choose_markup_kind(supported_markup_kinds)

    if shared_data:
        completion, data = shared_data
        return _resolve_completion(completion, data, preferred_markup_kind)
    return completion_item


def _sort_text(definition):
    """Ensure builtins appear at the bottom.
    Description is of format <type>: <module>.<item>
    """
    if definition.name.startswith("_"):
        # It's a 'hidden' func, put it next last
        return "z" + definition.name
    if definition.scope == "builtin":
        return "y" + definition.name

    # Else put it at the front
    return "a" + definition.name


def _kind(d):
    """Return the LSP type"""
    MAP = {
        "none": lsp.CompletionItemKind.Value,
        "type": lsp.CompletionItemKind.Class,
        "tuple": lsp.CompletionItemKind.Class,
        "dict": lsp.CompletionItemKind.Class,
        "dictionary": lsp.CompletionItemKind.Class,
        "function": lsp.CompletionItemKind.Function,
        "lambda": lsp.CompletionItemKind.Function,
        "generator": lsp.CompletionItemKind.Function,
        "class": lsp.CompletionItemKind.Class,
        "instance": lsp.CompletionItemKind.Reference,
        "method": lsp.CompletionItemKind.Method,
        "builtin": lsp.CompletionItemKind.Class,
        "builtinfunction": lsp.CompletionItemKind.Function,
        "module": lsp.CompletionItemKind.Module,
        "file": lsp.CompletionItemKind.File,
        "xrange": lsp.CompletionItemKind.Class,
        "slice": lsp.CompletionItemKind.Class,
        "traceback": lsp.CompletionItemKind.Class,
        "frame": lsp.CompletionItemKind.Class,
        "buffer": lsp.CompletionItemKind.Class,
        "dictproxy": lsp.CompletionItemKind.Class,
        "funcdef": lsp.CompletionItemKind.Function,
        "property": lsp.CompletionItemKind.Property,
        "import": lsp.CompletionItemKind.Module,
        "keyword": lsp.CompletionItemKind.Keyword,
        "constant": lsp.CompletionItemKind.Variable,
        "variable": lsp.CompletionItemKind.Variable,
        "value": lsp.CompletionItemKind.Value,
        "param": lsp.CompletionItemKind.Variable,
        "statement": lsp.CompletionItemKind.Keyword,
    }

    return MAP.get(d.type)



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging
import os

from pylsp._utils import find_parents

from .source import ConfigSource

log = logging.getLogger(__name__)

CONFIG_KEY = "flake8"
PROJECT_CONFIGS = [".flake8", "setup.cfg", "tox.ini"]

OPTIONS = [
    # mccabe
    ("max-complexity", "plugins.mccabe.threshold", int),
    # pycodestyle
    ("exclude", "plugins.pycodestyle.exclude", list),
    ("filename", "plugins.pycodestyle.filename", list),
    ("hang-closing", "plugins.pycodestyle.hangClosing", bool),
    ("ignore", "plugins.pycodestyle.ignore", list),
    ("max-line-length", "plugins.pycodestyle.maxLineLength", int),
    ("indent-size", "plugins.pycodestyle.indentSize", int),
    ("select", "plugins.pycodestyle.select", list),
    # flake8
    ("exclude", "plugins.flake8.exclude", list),
    ("extend-ignore", "plugins.flake8.extendIgnore", list),
    ("extend-select", "plugins.flake8.extendSelect", list),
    ("filename", "plugins.flake8.filename", list),
    ("hang-closing", "plugins.flake8.hangClosing", bool),
    ("ignore", "plugins.flake8.ignore", list),
    ("max-complexity", "plugins.flake8.maxComplexity", int),
    ("max-line-length", "plugins.flake8.maxLineLength", int),
    ("indent-size", "plugins.flake8.indentSize", int),
    ("select", "plugins.flake8.select", list),
    ("per-file-ignores", "plugins.flake8.perFileIgnores", list),
]


class Flake8Config(ConfigSource):
    """Parse flake8 configurations."""

    def user_config(self):
        config_file = self._user_config_file()
        config = self.read_config_from_files([config_file])
        return self.parse_config(config, CONFIG_KEY, OPTIONS)

    def _user_config_file(self):
        if self.is_windows:
            return os.path.expanduser("~\\.flake8")
        return os.path.join(self.xdg_home, "flake8")

    def project_config(self, document_path):
        files = find_parents(self.root_path, document_path, PROJECT_CONFIGS)
        config = self.read_config_from_files(files)
        return self.parse_config(config, CONFIG_KEY, OPTIONS)

    @classmethod
    def _parse_list_opt(cls, string):
        if string.startswith("\n"):
            return [s.strip().rstrip(",") for s in string.split("\n") if s.strip()]
        return [s.strip() for s in string.split(",") if s.strip()]



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import argparse
import logging
import logging.config
import sys
import time

try:
    import ujson as json
except Exception:
    import json

from ._version import __version__
from .python_lsp import (
    PythonLSPServer,
    start_io_lang_server,
    start_tcp_lang_server,
    start_ws_lang_server,
)

LOG_FORMAT = "%(asctime)s {} - %(levelname)s - %(name)s - %(message)s".format(
    time.localtime().tm_zone
)


def add_arguments(parser) -> None:
    parser.description = "Python Language Server"

    parser.add_argument(
        "--tcp", action="store_true", help="Use TCP server instead of stdio"
    )
    parser.add_argument(
        "--ws", action="store_true", help="Use Web Sockets server instead of stdio"
    )
    parser.add_argument("--host", default="127.0.0.1", help="Bind to this address")
    parser.add_argument("--port", type=int, default=2087, help="Bind to this port")
    parser.add_argument(
        "--check-parent-process",
        action="store_true",
        help="Check whether parent process is still alive using os.kill(ppid, 0) "
        "and auto shut down language server process when parent process is not alive. "
        "Note that this may not work on a Windows machine.",
    )

    log_group = parser.add_mutually_exclusive_group()
    log_group.add_argument(
        "--log-config", help="Path to a JSON file containing Python logging config."
    )
    log_group.add_argument(
        "--log-file",
        help="Redirect logs to the given file instead of writing to stderr. "
        "Has no effect if used with --log-config.",
    )

    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase verbosity of log output, overrides log config file",
    )

    parser.add_argument(
        "-V", "--version", action="version", version="%(prog)s v" + __version__
    )


def main() -> None:
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    args = parser.parse_args()
    _configure_logger(args.verbose, args.log_config, args.log_file)

    if args.tcp:
        start_tcp_lang_server(
            args.host, args.port, args.check_parent_process, PythonLSPServer
        )
    elif args.ws:
        start_ws_lang_server(args.port, args.check_parent_process, PythonLSPServer)
    else:
        stdin, stdout = _binary_stdio()
        start_io_lang_server(stdin, stdout, args.check_parent_process, PythonLSPServer)


def _binary_stdio():
    """Construct binary stdio streams (not text mode).

    This seems to be different for Window/Unix Python2/3, so going by:
        https://stackoverflow.com/questions/2850893/reading-binary-data-from-stdin
    """
    stdin, stdout = sys.stdin.buffer, sys.stdout.buffer
    return stdin, stdout


def _configure_logger(verbose=0, log_config=None, log_file=None) -> None:
    root_logger = logging.root

    if log_config:
        with open(log_config, encoding="utf-8") as f:
            logging.config.dictConfig(json.load(f))
    else:
        formatter = logging.Formatter(LOG_FORMAT)
        if log_file:
            log_handler = logging.handlers.RotatingFileHandler(
                log_file,
                mode="a",
                maxBytes=50 * 1024 * 1024,
                backupCount=10,
                encoding=None,
                delay=0,
            )
        else:
            log_handler = logging.StreamHandler()
        log_handler.setFormatter(formatter)
        root_logger.addHandler(log_handler)

    if verbose == 0:
        level = logging.WARNING
    elif verbose == 1:
        level = logging.INFO
    elif verbose >= 2:
        level = logging.DEBUG

    root_logger.setLevel(level)


if __name__ == "__main__":
    main()



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging
import re
from pathlib import Path

from pylsp import hookimpl
from pylsp.lsp import SymbolKind

log = logging.getLogger(__name__)


@hookimpl
def pylsp_document_symbols(config, document):
    symbols_settings = config.plugin_settings("jedi_symbols")
    all_scopes = symbols_settings.get("all_scopes", True)
    add_import_symbols = symbols_settings.get("include_import_symbols", True)
    definitions = document.jedi_names(all_scopes=all_scopes)
    symbols = []
    exclude = set({})
    redefinitions = {}
    pattern_import = re.compile(
        r"^\s*(?!#)\s*(from\s+[.\w]+(\.[\w]+)*\s+import\s+[\w\s,()*]+|import\s+[\w\s,.*]+)"
    )

    while definitions != []:
        d = definitions.pop(0)

        # Skip symbols imported from other modules.
        if not add_import_symbols:
            # Skip if there's an import in the code the symbol is defined.
            code = d.get_line_code()

            if pattern_import.match(code):
                continue

            # Skip imported symbols comparing module names.
            sym_full_name = d.full_name
            if sym_full_name is not None:
                document_dot_path = document.dot_path

                # We assume a symbol is imported from another module to start
                # with.
                imported_symbol = True

                # The last element of sym_full_name is the symbol itself, so
                # we need to discard it to do module comparisons below.
                if "." in sym_full_name:
                    sym_module_name = sym_full_name.rpartition(".")[0]
                else:
                    sym_module_name = sym_full_name

                # This is necessary to display symbols in init files (the checks
                # below fail without it).
                if document_dot_path.endswith("__init__"):
                    document_dot_path = document_dot_path.rpartition(".")[0]

                # document_dot_path is the module where the symbol is imported,
                # whereas sym_module_name is the one where it was declared.
                if document_dot_path in sym_module_name:
                    # If document_dot_path is in sym_module_name, we can safely assume
                    # that the symbol was declared in the document.
                    imported_symbol = False
                elif sym_module_name.split(".")[0] in document_dot_path.split("."):
                    # If the first module in sym_module_name is one of the modules in
                    # document_dot_path, we need to check if sym_module_name starts
                    # with the modules in document_dot_path.
                    document_mods = document_dot_path.split(".")
                    for i in range(1, len(document_mods) + 1):
                        submod = ".".join(document_mods[-i:])
                        if sym_module_name.startswith(submod):
                            imported_symbol = False
                            break

                # When there's no __init__.py next to a file or in one of its
                # parents, the checks above fail. However, Jedi has a nice way
                # to tell if the symbol was declared in the same file: if
                # sym_module_name starts by __main__.
                if imported_symbol:
                    if not sym_module_name.startswith("__main__"):
                        continue
            else:
                # We need to skip symbols if their definition doesn't have `full_name` info, they
                # are detected as a definition, but their description (e.g. `class Foo`) doesn't
                # match the code where they're detected by Jedi. This happens for relative imports.
                if _include_def(d):
                    if d.description not in d.get_line_code():
                        continue
                else:
                    continue

        if _include_def(d) and Path(document.path) == Path(d.module_path):
            tuple_range = _tuple_range(d)
            if tuple_range in exclude:
                continue

            kind = redefinitions.get(tuple_range, None)
            if kind is not None:
                exclude |= {tuple_range}

            if d.type == "statement":
                if d.description.startswith("self"):
                    kind = "field"

            symbol = {
                "name": d.name,
                "containerName": _container(d),
                "location": {
                    "uri": document.uri,
                    "range": _range(d),
                },
                "kind": _kind(d) if kind is None else _SYMBOL_KIND_MAP[kind],
            }
            symbols.append(symbol)

            if d.type == "class":
                try:
                    defined_names = list(d.defined_names())
                    for method in defined_names:
                        if method.type == "function":
                            redefinitions[_tuple_range(method)] = "method"
                        elif method.type == "statement":
                            redefinitions[_tuple_range(method)] = "field"
                        else:
                            redefinitions[_tuple_range(method)] = method.type
                    definitions = list(defined_names) + definitions
                except Exception:
                    pass
    return symbols


def _include_def(definition):
    return (
        # Don't tend to include parameters as symbols
        definition.type != "param"
        and
        # Unused vars should also be skipped
        definition.name != "_"
        and _kind(definition) is not None
    )


def _container(definition):
    try:
        # Jedi sometimes fails here.
        parent = definition.parent()
        # Here we check that a grand-parent exists to avoid declaring symbols
        # as children of the module.
        if parent.parent():
            return parent.name
    except:
        return None

    return None


def _range(definition):
    # This gets us more accurate end position
    definition = definition._name.tree_name.get_definition()
    (start_line, start_column) = definition.start_pos
    (end_line, end_column) = definition.end_pos
    return {
        "start": {"line": start_line - 1, "character": start_column},
        "end": {"line": end_line - 1, "character": end_column},
    }


def _tuple_range(definition):
    definition = definition._name.tree_name.get_definition()
    return (definition.start_pos, definition.end_pos)


_SYMBOL_KIND_MAP = {
    "none": SymbolKind.Variable,
    "type": SymbolKind.Class,
    "tuple": SymbolKind.Class,
    "dict": SymbolKind.Class,
    "dictionary": SymbolKind.Class,
    "function": SymbolKind.Function,
    "lambda": SymbolKind.Function,
    "generator": SymbolKind.Function,
    "class": SymbolKind.Class,
    "instance": SymbolKind.Class,
    "method": SymbolKind.Method,
    "builtin": SymbolKind.Class,
    "builtinfunction": SymbolKind.Function,
    "module": SymbolKind.Module,
    "file": SymbolKind.File,
    "xrange": SymbolKind.Array,
    "slice": SymbolKind.Class,
    "traceback": SymbolKind.Class,
    "frame": SymbolKind.Class,
    "buffer": SymbolKind.Array,
    "dictproxy": SymbolKind.Class,
    "funcdef": SymbolKind.Function,
    "property": SymbolKind.Property,
    "import": SymbolKind.Module,
    "keyword": SymbolKind.Variable,
    "constant": SymbolKind.Constant,
    "variable": SymbolKind.Variable,
    "value": SymbolKind.Variable,
    "param": SymbolKind.Variable,
    "statement": SymbolKind.Variable,
    "boolean": SymbolKind.Boolean,
    "int": SymbolKind.Number,
    "longlean": SymbolKind.Number,
    "float": SymbolKind.Number,
    "complex": SymbolKind.Number,
    "string": SymbolKind.String,
    "unicode": SymbolKind.String,
    "list": SymbolKind.Array,
    "field": SymbolKind.Field,
}


def _kind(d):
    """Return the VSCode Symbol Type"""
    return _SYMBOL_KIND_MAP.get(d.type)



# Copyright 2017-2020 Palantir Technologies, Inc.
# Copyright 2021- Python Language Server Contributors.

import logging

from pylsp import _utils, hookimpl, lsp

log = logging.getLogger(__name__)


@hookimpl
def pylsp_document_highlight(document, position):
    code_position = _utils.position_to_jedi_linecolumn(document, position)
    usages = document.jedi_script().get_references(**code_position)

    def is_valid(definition):
        return definition.line is not None and definition.column is not None

    def local_to_document(definition):
        return (
            not definition.module_path or str(definition.module_path) == document.path
        )

    return [
        {
            "range": {
                "start": {"line": d.line - 1, "character": d.column},
                "end": {"line": d.line - 1, "character": d.column + len(d.name)},
            },
            "kind": lsp.DocumentHighlightKind.Write
            if d.is_definition()
            else lsp.DocumentHighlightKind.Read,
        }
        for d in usages
        if is_valid(d) and local_to_document(d)
    ]
